#!/usr/bin/env python
#
# overlay.py - Functions for painting things on top of wx widgets.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains convenience functions for drawing arbitrary things
on top of ``wx`` widgets.

.. autosummary::
   :nosignatures:

   textOverlay
"""


import wx


def textOverlay(target,
                text,
                box=True,
                bgColour=None,
                fgColour=None):
    """Shows the given ``text`` on the given ``target``.

    :arg target:   ``wx.Window`` object

    :arg text:     Text to display

    :arg box:      If ``True`` (the default), the text is displayed in a
                   box.

    :arg bgColour: If ``box is True``, the box is filled with this colour.

    :arg fgColour: Colour to draw the text in
    """

    if bgColour is None:
        bgColour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_INFOBK)
    if fgColour is None:
        fgColour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_INFOTEXT)

    dc     = wx.ClientDC(target)
    w, h   = dc.GetSize().Get()
    w      = dc.DeviceToLogicalX(w)
    h      = dc.DeviceToLogicalY(h)
    tw, th = dc.GetTextExtent(text).Get()
    x      = int((w - tw) / 2)
    y      = int((h - th) / 2)

    if box:
        dc.SetBrush(wx.Brush(wx.Colour(bgColour)))
        dc.DrawRoundedRectangle(x - 5, y - 5, tw + 10, th + 10, 2)

    dc.SetTextForeground(wx.Colour(fgColour))
    dc.DrawText(text, x, y)



#!/usr/bin/env python
#
# test_togglepanel.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import fsleyes_widgets.togglepanel as togglepanel

import wx

from . import run_with_wx, simclick, realYield


def test_usage():
    run_with_wx(_test_usage)
def _test_usage():

    frame = wx.GetApp().GetTopWindow()
    pt = togglepanel.TogglePanel(frame, toggleSide=wx.TOP)
    pb = togglepanel.TogglePanel(frame, toggleSide=wx.BOTTOM)
    pl = togglepanel.TogglePanel(frame, toggleSide=wx.LEFT)
    pr = togglepanel.TogglePanel(frame, toggleSide=wx.RIGHT)

    pe = togglepanel.TogglePanel(frame, initialState=True)
    pc = togglepanel.TogglePanel(frame, initialState=False)

    assert pe.IsExpanded()
    assert not pc.IsExpanded()

    pe.Expand()
    assert pe.IsExpanded()

    pe.Collapse()
    assert not pe.IsExpanded()

    pe.Expand(False)
    assert not pe.IsExpanded()

    pe.Expand()
    assert pe.IsExpanded()

    pe.Toggle()
    assert not pe.IsExpanded()

    pe.Toggle()
    assert pe.IsExpanded()

    pnl = togglepanel.TogglePanel(frame)
    pl  = togglepanel.TogglePanel(frame, label='blag')

    assert pnl.GetLabel() is  None
    assert pl .GetLabel() == 'blag'

    pl.SetLabel('fum')
    assert pl .GetLabel() == 'fum'

    pl.SetLabel(None)
    assert pl .GetLabel() is None


def test_event():
    run_with_wx(_test_event)

def _test_event():

    event = [None]
    def ontog(ev):
        event[0] = (ev.GetEventObject(), ev.newState)

    frame   = wx.GetApp().GetTopWindow()
    fsizer  = wx.BoxSizer(wx.HORIZONTAL)
    tp      = togglepanel.TogglePanel(frame)
    panel   = tp.GetPane()
    tpsizer = wx.BoxSizer(wx.VERTICAL)
    content = wx.Button(panel, label='Hey')

    tpsizer.Add(content, flag=wx.EXPAND)
    panel.SetSizer(tpsizer)

    fsizer.Add(tp)
    frame.SetSizer(fsizer)
    frame.Layout()

    realYield()

    tp.Bind(togglepanel.EVT_TOGGLEPANEL_EVENT, ontog)

    sim = wx.UIActionSimulator()

    assert tp.IsExpanded()

    simclick(sim, tp.button)

    assert not tp.IsExpanded()

    assert event[0] is not None
    assert event[0][0] is tp
    assert not event[0][1]

    event[0] = None
    simclick(sim, tp.button)

    assert tp.IsExpanded()

    assert event[0] is not None
    assert event[0][0] is tp
    assert event[0][1]



#!/usr/bin/env python
#
# test_floatspin.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import time

import wx
from unittest import mock

import fsleyes_widgets.floatspin as floatspin

from .test_floatslider import (_test_widget_logic,
                               _test_widget_logic_integer,
                               _test_widget_changeRange)

from . import run_with_wx, simclick, simtext, simkey, simfocus, realYield, addall



def test_FloatSpinCtrl_logic():
    run_with_wx(_test_FloatSpinCtrl_logic)
def _test_FloatSpinCtrl_logic():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame)
    _test_widget_logic(spin)


def test_FloatSpinCtrl_logic_integer():
    run_with_wx(_test_FloatSpinCtrl_logic_integer)
def _test_FloatSpinCtrl_logic_integer():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame, style=floatspin.FSC_INTEGER)
    _test_widget_logic_integer(spin)


def test_FloatSpinCtrl_changeRange():
    run_with_wx(_test_FloatSpinCtrl_changeRange)
def _test_FloatSpinCtrl_changeRange():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame)
    _test_widget_changeRange(spin)


def test_FloatSpinCtrl_nolimit():
    run_with_wx(_test_FloatSpinCtrl_nolimit)
def _test_FloatSpinCtrl_nolimit():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame, style=floatspin.FSC_NO_LIMIT)
    spin.SetRange(0, 100)

    values = [-100, -50, -1, 0, 1, 50, 99, 100, 101, 150, 200]


    for v in values:
        spin.SetValue(v)
        assert spin.GetValue() == v


def test_FloatSpinCtrl_spinButtons():
    run_with_wx(_test_FloatSpinCtrl_spinButtons)
def _test_FloatSpinCtrl_spinButtons():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame,
                                    value=0,
                                    minValue=0,
                                    maxValue=100,
                                    increment=1,
                                    evDelta=0.01)

    ev = mock.MagicMock()

    # The FloatSpinCtrl class throttles the
    # rate at which spin button events are
    # processed, so we have to wait between
    # triggering events.

    time.sleep(0.05)
    spin._FloatSpinCtrl__onSpinDown(ev)
    assert spin.GetValue() == 0
    time.sleep(0.05)
    spin._FloatSpinCtrl__onSpinUp(ev)
    assert spin.GetValue() == 1
    time.sleep(0.05)
    spin._FloatSpinCtrl__onSpinDown(ev)
    assert spin.GetValue() == 0
    spin.SetValue(100)
    time.sleep(0.05)
    spin._FloatSpinCtrl__onSpinUp(ev)
    assert spin.GetValue() == 100

    results = [None]

    def handler(ev):
        results[0] = ev.value

    spin.Bind(floatspin.EVT_FLOATSPIN, handler)

    time.sleep(0.05)
    spin.SetValue(50)
    spin._FloatSpinCtrl__onSpinUp(ev)
    realYield()
    assert results[0] == 51

    results[0] = None
    time.sleep(0.05)
    spin.SetValue(50)
    spin._FloatSpinCtrl__onSpinDown(ev)
    realYield()
    assert results[0] == 49


# Mouse wheel down/up
def test_FloatSpinCtrl_mouseWheel():
    run_with_wx(_test_FloatSpinCtrl_mouseWheel)
def _test_FloatSpinCtrl_mouseWheel():

    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame, style=floatspin.FSC_MOUSEWHEEL)
    spin.SetRange(0, 100)

    ev = mock.MagicMock()

    testcases = [
        ( 1, wx.MOUSE_WHEEL_VERTICAL,    1),
        (-1, wx.MOUSE_WHEEL_VERTICAL,   -1),
        ( 1, wx.MOUSE_WHEEL_HORIZONTAL, -1),
        (-1, wx.MOUSE_WHEEL_HORIZONTAL,  1)]

    results = [None]

    def handler(ev):
        results[0] = ev.value

    spin.Bind(floatspin.EVT_FLOATSPIN, handler)


    for rot, axis, expected in testcases:

        results[0] = None
        time.sleep(0.05)
        spin.SetValue(50)

        ev.GetWheelAxis.return_value     = axis
        ev.GetWheelRotation.return_value = rot

        spin._FloatSpinCtrl__onMouseWheel(ev)
        realYield()

        assert results[0]      == 50 + expected
        assert spin.GetValue() == 50 + expected

def test_FloatSpinCtrl_text():
    run_with_wx(_test_FloatSpinCtrl_text)
def _test_FloatSpinCtrl_text():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame)
    sim   = wx.UIActionSimulator()

    addall(frame, [spin])

    spin.SetRange(0, 100)

    result = [0]

    def handler(ev):
        result[0] = ev.value

    enter = [False]

    def enterHandler(ev):
        enter[0] = True

    spin.Bind(floatspin.EVT_FLOATSPIN, handler)
    spin.Bind(wx.EVT_TEXT_ENTER,       enterHandler)

    # input, expected
    testcases = [
        (   '',          0),
        (   '1',         1),
        (   '5',         5),
        ('Baba',         5),
        (   '0',         0),
        (  '20',        20),
        ('-423',         0),
        (  '99',        99),
        ('1055',       100),
        (  '25',        25),
        (  '25.42',  25.42),
        ('1E2',        100),
        ('0.5e2',       50),
        ('2e-1',       0.2),
        ('0.1e-2',   0.001),
    ]

    realYield()

    for text, expected in testcases:

        oldValue  = spin.GetValue()
        result[0] = None
        enter[0]  = False
        spin.textCtrl.ChangeValue(text)
        spin._FloatSpinCtrl__onText(None)
        realYield()
        assert spin.GetValue() == expected

        try:
            float(text)
            assert enter[0]
        except:
            assert not enter[0]

        if oldValue == expected: assert result[0] is None
        else:                    assert result[0] == expected


def test_FloatSpinCtrl_KeyDownUp():
    run_with_wx(_test_FloatSpinCtrl_KeyDownUp)
def _test_FloatSpinCtrl_KeyDownUp():
    frame = wx.GetApp().GetTopWindow()
    spin  = floatspin.FloatSpinCtrl(frame)
    sim   = wx.UIActionSimulator()

    spin.SetRange(0, 100)

    result = [0]

    def handler(ev):
        result[0] = ev.value

    spin.Bind(floatspin.EVT_FLOATSPIN, handler)

    # initialValue, key, expected
    testcases = [
        (50,  wx.WXK_UP,     51),
        (50,  wx.WXK_DOWN,   49),
        (1,   wx.WXK_UP,      2),
        (1,   wx.WXK_DOWN,    0),
        (0,   wx.WXK_UP,      1),
        (0,   wx.WXK_DOWN,    0),
        (99,  wx.WXK_UP,    100),
        (99,  wx.WXK_DOWN,   98),
        (100, wx.WXK_UP,    100),
        (100, wx.WXK_DOWN,   99),
    ]

    spin.textCtrl.SetFocus()

    realYield()

    for value, key, expected in testcases:
        result[0] = None

        spin.SetValue(value)
        simkey(sim, spin.textCtrl, key)

        assert spin.GetValue() == expected

        if expected != value: assert result[0] == expected
        else:                 assert result[0] is None


def test_FloatSpinCtrl_TextLoseFocus():
    run_with_wx(_test_FloatSpinCtrl_TextLoseFocus)
def _test_FloatSpinCtrl_TextLoseFocus():
    frame = wx.GetApp().GetTopWindow()
    sim   = wx.UIActionSimulator()
    spin  = floatspin.FloatSpinCtrl(frame)
    dummy = wx.TextCtrl(frame)

    spin.SetRange(0, 100)

    result = [None]

    def handler(ev):
        result[0] = ev.value

    spin.Bind(floatspin.EVT_FLOATSPIN, handler)

    testcases = [
        (   '1',         1),
        (   '5',         5),
        ('Baba',         5),
        (   '0',         0),
        (  '20',        20),
        ('-423',         0),
        (  '99',        99),
        ('1055',       100),
        (  '25',        25),
        (  '25.42',  25.42),
    ]

    for text, expected in testcases:

        oldValue = spin.GetValue()

        result[0] = None
        spin.textCtrl.SetFocus()
        simtext(sim, spin.textCtrl, text, enter=False)
        simfocus(spin, dummy)

        assert spin.GetValue() == expected

        if oldValue == expected: assert result[0] is None
        else:                    assert result[0] == expected



#!/usr/bin/env python
#
# test_isalive.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

from unittest import mock
import sys

import wx

import fsleyes_widgets       as fw
import fsleyes_widgets.utils as fwutils

from . import run_with_wx, run_without_wx, realYield


def test_isalive():
    run_with_wx(_test_isalive)

def _test_isalive():
    frame   = wx.GetApp().GetTopWindow()
    sizer   = wx.BoxSizer(wx.VERTICAL)
    panel   = wx.Panel(frame)
    child1  = wx.Panel(panel)
    child2  = wx.Panel(panel)

    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)

    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(child1, flag=wx.EXPAND, proportion=1)
    sizer.Add(child2, flag=wx.EXPAND, proportion=1)
    panel.SetSizer(sizer)

    assert fw.isalive(child1)
    assert fw.isalive(child2)

    sizer.Detach(child1)
    child1.Destroy()

    sizer.Layout()
    frame.Refresh()

    realYield(100)

    assert not fw.isalive(child1)
    assert     fw.isalive(child2)


def test_wxversion():
    fw.wxversion()
    with run_without_wx():
        assert fw.wxversion() is None


def test_wxVersion():
    with mock.patch('wx.__version__', '4.0.0'):
        assert fw.wxVersion() == '4.0.0'
    with mock.patch('wx.__version__', '4.1.1'):
        assert fw.wxVersion() == '4.1.1'
    with mock.patch('wx.__version__', '4.0.7.post2'):
        assert fw.wxVersion() == '4.0.7'
    with mock.patch('wx.__version__', 'invalid'):
        assert fw.wxVersion() is None
    with run_without_wx():
        assert fw.wxVersion() is None


def test_wxFlavour():
    with run_without_wx():
        assert fw.wxFlavour() is None
    with mock.patch('wx.PlatformInfo', ['phoenix']):
        assert fw.wxFlavour()  == fw.WX_PHOENIX
    with mock.patch('wx.PlatformInfo', ['oldversion']):
        assert fw.wxFlavour()  == fw.WX_PYTHON


def test_wxPlatform():
    with run_without_wx():
        assert fw.wxPlatform() is None
    with mock.patch('wx.PlatformInfo', ['cocoa']):
        assert fw.wxPlatform()  == fw.WX_MAC_COCOA
    with mock.patch('wx.PlatformInfo', ['carbon']):
        assert fw.wxPlatform()  == fw.WX_MAC_CARBON

    with mock.patch('wx.PlatformInfo', ['gtk3']):
        assert fw.wxPlatform()  == fw.WX_GTK3
    with mock.patch('wx.PlatformInfo', ['gtk2']):
        assert fw.wxPlatform()  == fw.WX_GTK2
    # PlatformInfo should have "gtk2" or "gtk3";
    # if for some reason it only has "gtk",
    # wxPlatform will assume gtk2
    with mock.patch('wx.PlatformInfo', ['gtk']):
        assert fw.wxPlatform()  == fw.WX_GTK2
    with mock.patch('wx.PlatformInfo', ['atari']):
        assert fw.wxPlatform()  == fw.WX_UNKNOWN


def test_frozen():
    assert not fw.frozen()
    with mock.patch('fsleyes_widgets.utils.sys',
                    mock.Mock(wraps=sys)) as mock_sys:
        mock_sys.frozen = True
        assert fw.frozen()
        mock_sys.frozen = False
        assert not fw.frozen()


def test_canHaveGui():
    with run_without_wx():
        fw.canHaveGui.cache_clear()
        assert not fw.canHaveGui()
    with mock.patch('wx.App.IsDisplayAvailable', return_value=False):
        fw.canHaveGui.cache_clear()
        assert not fw.canHaveGui()
    with mock.patch('wx.App.IsDisplayAvailable', return_value=True):
        fw.canHaveGui.cache_clear()
        assert fw.canHaveGui()
    fw.canHaveGui.cache_clear()


@run_with_wx
def test_haveGui1():
    assert fw.haveGui()
    fw.canHaveGui.cache_clear()


def test_haveGui2():
    with run_without_wx():
        assert not fw.haveGui()
        fw.canHaveGui.cache_clear()

    assert not fw.haveGui()
    fw.canHaveGui.cache_clear()


def test_inSSHSession():

    sshVars = ['SSH_CLIENT', 'SSH_TTY']

    for sv in sshVars:
        with mock.patch.dict('os.environ', **{sv : '1'}):
            assert fw.inSSHSession()

    with mock.patch('os.environ', {}):
        assert not fw.inSSHSession()


def test_inVNCSession():
    vncVars = ['VNCDESKTOP', 'X2GO_SESSION', 'NXSESSIONID']

    for vv in vncVars:
        with mock.patch.dict('os.environ', **{vv : '1'}):
            assert fw.inVNCSession()

    with mock.patch('os.environ', {}):
        assert not fw.inVNCSession()



#!/usr/bin/env python
#
# test_numberdialog.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import wx

import fsleyes_widgets.numberdialog as numdlg

from . import run_with_wx, simclick, simtext, simkey, realYield


def test_NumberDialog_create():
    run_with_wx(_test_NumberDialog_create)
def _test_NumberDialog_create():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()

    testcases = [
        ({'real' : False, 'initial' :    999}, 'ok',        999),
        ({'real' : False, 'initial' :   -999}, 'ok',       -999),
        ({'real' : True,  'initial' : 0.3443}, 'ok',     0.3443),
        ({'real' : True,  'initial' :  -0.48}, 'ok',      -0.48),
        ({'real' : True,  'initial' : 0.3443}, 'cancel',   None),
    ]

    for kwargs, target, expected in testcases:
        dlg = numdlg.NumberDialog(
            frame,
            title='NumberDialog test',
            message='Enter a number',
            **kwargs)

        if target == 'ok': target = dlg.okButton
        else:              target = dlg.cancelButton

        dlg.Show()
        realYield()

        simclick(sim, target)
        assert dlg.GetValue() == expected


def simtextnd(nd, text, changed):
    class FakeEv:
        def __init__(self, changed):
            self.changed = changed
    nd.floatSpinCtrl.textCtrl.ChangeValue(text)
    nd.floatSpinCtrl._FloatSpinCtrl__onText(None)
    nd._NumberDialog__onEnter(FakeEv(changed))


def test_NumberDialog_limit():
    run_with_wx(_test_NumberDialog_limit)
def _test_NumberDialog_limit():
    frame = wx.GetApp().GetTopWindow()

    # kwargs, input, needClick, expected
    testcases = [
        ({'real'     : False,
          'minValue' : 0,
          'maxValue' : 100}, '50', False, 50),
        ({'real'     : False,
          'minValue' : 0,
          'maxValue' : 100}, '0',  False,  0),
        ({'real'     : False,
          'minValue' : 0,
          'maxValue' : 100}, '-1', True,  0),
        ({'real'     : False,
          'minValue' : 0,
          'maxValue' : 100}, '100', False, 100),
        ({'real'     : False,
          'minValue' : 0,
          'maxValue' : 100}, '101', True, 100),

        ({'real'     : True,
          'minValue' : 0,
          'maxValue' : 1}, '0.745', False, 0.745),
        ({'real'     : True,
          'minValue' : 0,
          'maxValue' : 1}, '0.0', False, 0.0),
        ({'real'     : True,
          'minValue' : 0,
          'maxValue' : 1}, '-25.9', True, 0.0),
        ({'real'     : True,
          'minValue' : 0,
          'maxValue' : 1}, '25.9', True, 1.0),
        ({'real'     : True,
          'minValue' : 0,
          'maxValue' : 1}, '1.0', False, 1.0),
    ]

    for kwargs, text, needClick, expected in testcases:
        dlg = numdlg.NumberDialog(frame, **kwargs)

        dlg.Show()

        simtextnd(dlg, text, needClick)
        if needClick:
            assert dlg.GetValue() is None
            dlg._NumberDialog__onOk(None)

        assert dlg.GetValue() == expected



#!/usr/bin/env python
#
# test_imagepanel.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path as op

import wx

from . import run_with_wx, realYield

import fsleyes_widgets.imagepanel as ip


datadir = op.join(op.dirname(__file__), 'testdata', 'bitmapbuttons')


def test_ImagePanel():
    run_with_wx(_test_ImagePanel)
def _test_ImagePanel():

    icon = op.join(datadir, 'true.png')
    icon = wx.Bitmap(icon,  wx.BITMAP_TYPE_PNG)
    icon = icon.ConvertToImage()

    frame  = wx.GetApp().GetTopWindow()
    panel1 = ip.ImagePanel(frame)
    realYield()
    panel1.SetImage(icon)
    realYield()
    panel2 = ip.ImagePanel(frame, preserveAspect=True)
    realYield()
    panel2.SetImage(icon)
    realYield()



#!/usr/bin/env python
#
# placeholder_textctrl.py - The PlaceholderTextCtrl class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlaceholderTextCtrl` which is a
``wx.TextCtrl`` that displays some placeholder text when it is empty
and unfocused.
"""


import wx


class PlaceholderTextCtrl(wx.TextCtrl):
    """the :class:`PlaceholderTextCtrl` is a ``wx.TextCtrl`` that displays
    some placeholder text when it is empty and unfocused.

    I wrote this class as a substitute for the ``wx.SearchCtrl`` which, under
    OSX Cocoa (wxPython 3.0.2.0) has focusing bugs that make it unusable.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``PlaceholderTextCtrl``.

        :arg placeholder:       The text to display when the control is empty
                                and unfocused.

        :arg placeholderColour: Colour to display the placeholder text in.

        All other arguments are passed to ``wx.TextCtrl.__init__``.
        """

        placeholder       = str(kwargs.pop('placeholder', ''))
        placeholderColour =     kwargs.pop('placeholderColour',
                                           (150, 150, 150))

        self.__fgColour = (0, 0, 0)

        wx.TextCtrl.__init__(self, *args, **kwargs)

        self.SetPlaceholder(      placeholder)
        self.SetPlaceholderColour(placeholderColour)

        self.Bind(wx.EVT_SET_FOCUS,  self.__onSetFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.__onKillFocus)

        self.__onKillFocus(None)


    def GetPlaceholder(self):
        """Returns the place holder text. """
        return self.__placeholder


    def GetPlaceholderColour(self):
        """Returns the placeholder text colour."""
        return self.__placeholderColour


    def SetPlaceholder(self, placeholder):
        """Sets the placeholder text. """

        if placeholder is None:
            placeholder = ''

        self.__placeholder = str(placeholder)


    def SetPlaceholderColour(self, colour):
        """Sets the placeholder text colour. """

        colour = list(colour)

        if len(colour) != 3:
            raise ValueError('Colour must be an RGB sequence')

        for i, c in enumerate(colour):

            if c < 0:   c = 0
            if c > 255: c = 255

            colour[i] = c

        self.__placeholderColour = colour


    def GetValue(self):
        """Returns the value currently contained in this
        ``PlaceholderTextCtrl``.
        """
        value = wx.TextCtrl.GetValue(self)

        if value == self.__placeholder: return ''
        else:                           return value


    def SetValue(self, value):
        """Set the value in this ``PlaceholderTextCtrl``. """
        wx.TextCtrl.SetValue(self, value)
        self.SetForegroundColour(self.__fgColour)


    def __onSetFocus(self, ev):
        """Called when this ``PlaceholderTextCtrl`` gains focus.

        Clears the placeholder text if necessary.
        """
        ev.Skip()

        value = wx.TextCtrl.GetValue(self)

        if value == self.__placeholder:

            self.SetForegroundColour(self.__fgColour)
            self.SetValue('')


    def __onKillFocus(self, ev):
        """Called when this ``PlaceholderTextCtrl`` loses focus.

        Displays the placeholder text if necessary.
        """
        if ev is not None:
            ev.Skip()

        value = wx.TextCtrl.GetValue(self)

        if value.strip() == '':
            self.__fgColour = self.GetForegroundColour()
            self.SetValue(           self.__placeholder)
            self.SetForegroundColour(self.__placeholderColour)



#!/usr/bin/env python

import wx
from unittest import mock
import numpy as np

from . import run_with_wx, simmove, realYield

import fsleyes_widgets.widgetgrid as widgetgrid


def test_create():
    run_with_wx(_test_create)
def _test_create():
    frame = wx.GetApp().GetTopWindow()
    grid = widgetgrid.WidgetGrid(frame)
    grid.SetGridSize(10, 10)
    grid.ShowRowLabels()
    grid.ShowColLabels()

    for i in range(10):
        grid.SetRowLabel(i, 'row {}'.format(i + 1))
        grid.SetColLabel(i, 'col {}'.format(i + 1))

    for i in range(10):
        for j in range(10):
            grid.SetText(i, j, 'cell [{}, {}]'.format(i + 1, j + 1))

    grid.Refresh()

    exprowlbls = ['row {}'.format(i + 1) for i in range(10)]
    expcollbls = ['col {}'.format(i + 1) for i in range(10)]

    assert grid.GetRowLabels() == exprowlbls
    assert grid.GetColLabels() == expcollbls

    for i in range(10):
        for j in range(10):
            exp = 'cell [{}, {}]'.format(i + 1, j + 1)
            assert grid.GetWidget(i, j).GetLabel() == exp



def test_reorder():
    run_with_wx(_test_reorder)
def _test_reorder():
    frame = wx.GetApp().GetTopWindow()
    grid = widgetgrid.WidgetGrid(frame)

    grid.SetGridSize(1, 10)

    labels = ['col {}' .format(i) for i in range(10)]
    cells  = ['cell {}'.format(i) for i in range(10)]

    grid.SetColLabels(labels)
    grid.SetRowLabels(['row'])

    for i in range(10): grid.SetText(0,  i, cells[i])

    grid.Refresh()
    realYield()
    for i in range(5):
        neworder = np.array(range(10))
        np.random.shuffle(neworder)
        grid.ReorderColumns(neworder)
        grid.Refresh()
        realYield()

        for i in range(10):

            assert grid.GetColLabel( i)            == labels[neworder[i]]
            assert grid.GetWidget(0, i).GetLabel() == cells[ neworder[i]]

        labels = [labels[i] for i in neworder]
        cells  = [cells[ i] for i in neworder]


class FakeMouseState(object):
    def __init__(self):
        self.pos = wx.Point(0, 0)
    def GetPosition(self):
        return self.pos

class FakeEV(object):

    def __init__(self, evo):
        self.evo = evo

    def Skip(self):
        pass

    def GetEventObject(self):
        return self.evo


def cxy(target, pos):
    w, h = target.GetClientSize().Get()
    x, y = target.GetScreenPosition()
    x += w * pos[0]
    y += h * pos[1]
    return int(x), int(y)





def test_reorder_events():
    run_with_wx(_test_reorder_events)
def _test_reorder_events():
    realYield()
    frame = wx.GetApp().GetTopWindow()
    frame.SetSize((800, 600))
    sim   = wx.UIActionSimulator()
    grid  = widgetgrid.WidgetGrid(frame, style=widgetgrid.WG_DRAGGABLE_COLUMNS)
    sizer = wx.BoxSizer(wx.VERTICAL)
    frame.SetSizer(sizer)
    sizer.Add(grid, flag=wx.EXPAND, proportion=1)

    grid.ShowColLabels()
    grid.SetGridSize(1, 5)

    labels = ['col {}' .format(i) for i in range(5)]
    cells  = ['cell {}'.format(i) for i in range(5)]

    for i in range(5): grid.SetColLabel(i, labels[i])
    for i in range(5): grid.SetText(0,  i, cells[i])

    grid.Refresh()
    frame.Layout()

    # (clicked column, drop column, drop pos, expected order)
    tests = [
        (0, 0, 0.25, [0, 1, 2, 3, 4]),
        (0, 0, 0.75, [0, 1, 2, 3, 4]),
        (1, 1, 0.25, [0, 1, 2, 3, 4]),
        (1, 1, 0.75, [0, 1, 2, 3, 4]),
        (0, 1, 0.25, [0, 1, 2, 3, 4]),
        (1, 2, 0.25, [0, 1, 2, 3, 4]),
        (1, 0, 0.75, [0, 1, 2, 3, 4]),
        (2, 1, 0.75, [0, 1, 2, 3, 4]),
        (0, 1, 0.75, [1, 0, 2, 3, 4]),
        (1, 2, 0.75, [0, 2, 1, 3, 4]),
        (1, 0, 0.25, [1, 0, 2, 3, 4]),
        (2, 1, 0.25, [0, 2, 1, 3, 4]),
        (0, 2, 0.25, [1, 0, 2, 3, 4]),
        (0, 2, 0.75, [1, 2, 0, 3, 4]),
        (2, 0, 0.25, [2, 0, 1, 3, 4]),
        (2, 0, 0.75, [0, 2, 1, 3, 4]),
        (1, 3, 0.25, [0, 2, 1, 3, 4]),
        (1, 3, 0.75, [0, 2, 3, 1, 4]),
        (3, 1, 0.25, [0, 3, 1, 2, 4]),
        (3, 1, 0.75, [0, 1, 3, 2, 4]),
        (1, 4, 0.25, [0, 2, 3, 1, 4]),
        (1, 4, 0.75, [0, 2, 3, 4, 1]),
    ]

    realYield()
    for clickcol, dropcol, droppos, exporder in tests:

        fakestate = FakeMouseState()

        with mock.patch('wx.GetMouseState', return_value=fakestate):

            cwidget = grid.colLabels[clickcol].GetParent()
            dwidget = grid.colLabels[dropcol] .GetParent()

            ev = FakeEV(cwidget)
            grid._WidgetGrid__onColumnLabelMouseDown(ev)
            realYield()

            fakestate.pos = wx.Point(*cxy(dwidget, (droppos, 0.5)))
            with mock.patch('wx.FindWindowAtPointer',
                            return_value=[dwidget]):
                grid._WidgetGrid__onColumnLabelMouseDrag(ev)
                realYield()
                ev.evo = dwidget
                grid._WidgetGrid__onColumnLabelMouseUp(ev)
                realYield()

            explabels = [labels[i] for i in exporder]
            gotlabels = grid.GetColLabels()

            assert explabels == gotlabels

            labels = gotlabels


def test_reorder_events_draglimit():
    run_with_wx(_test_reorder_events_draglimit)
def _test_reorder_events_draglimit():
    realYield()
    frame = wx.GetApp().GetTopWindow()
    frame.SetSize((800, 600))
    sim   = wx.UIActionSimulator()
    grid  = widgetgrid.WidgetGrid(frame, style=widgetgrid.WG_DRAGGABLE_COLUMNS)
    sizer = wx.BoxSizer(wx.VERTICAL)
    frame.SetSizer(sizer)
    sizer.Add(grid, flag=wx.EXPAND, proportion=1)

    grid.ShowColLabels()
    grid.SetGridSize(1, 5)
    grid.SetDragLimit(2)

    labels = ['col {}' .format(i) for i in range(5)]
    cells  = ['cell {}'.format(i) for i in range(5)]

    for i in range(5): grid.SetColLabel(i, labels[i])
    for i in range(5): grid.SetText(0,  i, cells[i])

    grid.Refresh()
    frame.Layout()

    # (clicked column, drop column, drop pos, expected order)
    tests = [
        (0, 1, 0.75, [1, 0, 2, 3, 4]),
        (1, 2, 0.75, [0, 2, 1, 3, 4]),
        (1, 0, 0.25, [1, 0, 2, 3, 4]),
        (2, 1, 0.25, [0, 2, 1, 3, 4]),

        # drop is past drag limit -> clamp
        (0, 3, 0.75, [1, 2, 0, 3, 4]),
        (0, 4, 0.75, [1, 2, 0, 3, 4]),

        # past drag limit -> no-ops
        (3, 1, 0.25, [0, 1, 2, 3, 4]),
        (3, 4, 0.75, [0, 1, 2, 3, 4]),
        (4, 3, 0.25, [0, 1, 2, 3, 4]),
    ]

    realYield()
    for clickcol, dropcol, droppos, exporder in tests:

        fakestate = FakeMouseState()

        with mock.patch('wx.GetMouseState', return_value=fakestate):

            cwidget = grid.colLabels[clickcol].GetParent()
            dwidget = grid.colLabels[dropcol] .GetParent()

            ev = FakeEV(cwidget)
            grid._WidgetGrid__onColumnLabelMouseDown(ev)
            realYield()

            fakestate.pos = wx.Point(*cxy(dwidget, (droppos, 0.5)))
            with mock.patch('wx.FindWindowAtPointer',
                            return_value=[dwidget]):
                grid._WidgetGrid__onColumnLabelMouseDrag(ev)
                realYield()
                ev.evo = dwidget
                grid._WidgetGrid__onColumnLabelMouseUp(ev)
                realYield()

            explabels = [labels[i] for i in exporder]
            gotlabels = grid.GetColLabels()

            assert explabels == gotlabels

            labels = gotlabels



#!/usr/bin/env python
#
# status.py - A simple interface for displaying messages.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This is a little module which provides an interface for displaying a
message, or status update, to the user.  The ``status`` module provides the
following functions:

 .. autosummary::
    :nosignatures:

    setTarget
    update
    clearStatus

A couple of other functions are also provided, for reporting error messages
to the user:

 .. autosummary::
    :nosignatures:

    reportError
    reportIfError
    reportErrorDecorator


The :func:`update` function may be used to display a message. By default, the
message is simply logged (via the ``logging`` module). However, if a status
target has been set via the :func:`setTarget` function, the message is also
passed to this target.


.. warning:: If the status update target is a ``wx`` GUI object, you must
             make sure that it is updated asynchronously (e.g. via
             ``wx.CallAfter``).
"""


import            threading
import            contextlib
import            logging
import            inspect
import os.path as op


log = logging.getLogger(__name__)


_statusUpdateTarget = None
"""A reference to the status update target - this is ``None`` by default, and
can be set via :func:`setTarget`.
"""


_clearThread = None
"""Reference to a :class:`ClearThread`, which is a daemon thread that clears
the status after the timeout passed to the :func:`update` function.
"""


def setTarget(target):
    """Set a target function to receive status updates. The ``target`` must
    be a function which accepts a string as its sole parameter.
    """
    global _statusUpdateTarget
    _statusUpdateTarget = target


def update(message, timeout=1.0):
    """Display a status update to the user. The message is logged and,
    if a status update target has been set, passed to the target.

    :arg timeout: Timeout (in seconds) after which the status will be
                  cleared (via the :class:`ClearThread`). Pass in ``None``
                  to disable this behaviour.


    .. note:: The ``timeout`` method only makes sense to use if the status
              target is a GUI widget of some sort.
    """

    global _clearThread
    global _statusUpdateTarget

    if log.getEffectiveLevel() == logging.DEBUG:

        frame   = inspect.stack()[1]
        module  = frame[1]
        linenum = frame[2]
        module  = op.basename(module)

        log.debug('[{}:{}] {}'.format(module, linenum, message))

    if _statusUpdateTarget is None:
        return

    try:
        _statusUpdateTarget(message)

    except Exception as e:
        log.warning('Target raised excepton {}'.format(e), exc_info=True)
        return

    if timeout is not None:
        log.debug('timeout is not None - starting clear thread')

        if _clearThread is None:
            _clearThread = ClearThread()
            _clearThread.start()

        _clearThread.clear(timeout)
    else:
        if _clearThread is not None:
            _clearThread.veto()
            log.debug('No timeout - vetoing clear thread')


def clearStatus():
    """Clear the status. If a status update target has been set, it is passed
    the empty string.
    """
    if _statusUpdateTarget is None:
        return

    _statusUpdateTarget('')


def reportError(title, msg, err):
    """Reports an error to the user in a generic manner. If a GUI is available,
    a ``wx.MessageBox`` is shown. Otherwise a log message is generated.
    """

    msg = '{}\n\nDetails: {}'.format(msg, str(err))

    try:
        import wx
        wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)

    except ImportError:
        log.error('{}: {}'.format(title, msg))


@contextlib.contextmanager
def reportIfError(title, msg, raiseError=True, report=True):
    """A context manager which calls :func:`reportError` if the enclosed code
    raises an ``Exception``.

    :arg raiseError: If ``True``, the ``Exception`` which was raised is
                     propagated upwards.

    :arg report:     Defaults to ``True``. If ``False``, an error message
                     is logged, but :func:`reportError` is not called.
    """
    try:
        yield

    except Exception as e:

        log.error('{}: {}'.format(title, msg))

        if report:
            reportError(title, msg, e)

        if raiseError:
            raise


def reportErrorDecorator(*args, **kwargs):
    """A decorator which wraps the decorated function with
    :func:`reportIfError`.
    """

    def decorator(func):
        def wrapper(*wargs, **wkwargs):
            with reportIfError(*args, **kwargs):
                func(*wargs, **wkwargs)

        return wrapper

    return decorator


class ClearThread(threading.Thread):
    """The ``ClearThread`` is a daemon thread used by the :func:`update`
    function. Only one ``ClearThread`` is ever started - it is started on the
    first call to ``update`` when a timeout is specified.

    The ``ClearThread`` waits until the :meth:`clear` method is called.
    It then waits for the specified timeout and, unless another call to
    :meth:`clear`, or a call to :meth:`veto` has been made, clears the
    status via a call to :func:`clearStatus`.
    """


    def __init__(self):
        """Create a ``ClearThread``. """

        threading.Thread.__init__(self)

        self.daemon       = True
        self.__clearEvent = threading.Event()
        self.__vetoEvent  = threading.Event()
        self.__dieEvent   = threading.Event()
        self.__timeout    = None


    def clear(self, timeout):
        """Clear the status after the specified timeout (in seconds). """

        self.__timeout = timeout
        self.__vetoEvent .clear()
        self.__clearEvent.set()


    def veto(self):
        """If this ``ClearThread`` is waiting on a timeout to clear
        the status, a call to ``veto`` will prevent it from doing so.
        """
        self.__vetoEvent.set()


    def die(self):
        """This method may be used to force the ``ClearThread`` to exit
        immediately, rather than when the process exits.


        To cleanly kill the ``ClearThread``, follow this procedure::

            import fsleyes_widgets.utils.status as status

            status._clearThread.die()
            status._clearThread.clear(0.1)
            status._clearThread.join()
            status._clearThread = None


        If the :func:`update` function is used again, a new ``ClearThread``
        will automatically be started.
        """
        self.__dieEvent.set()


    def run(self):
        """The ``ClearThread`` function. Infinite loop which waits until
        the :meth:`clear` method is called, and then clears the status
        (via a call to :func:`clearStatus`).
        """

        while True:

            self.__vetoEvent .clear()
            self.__clearEvent.wait()
            self.__clearEvent.clear()

            if self.__dieEvent.is_set():
                return

            # http://bugs.python.org/issue14623
            #
            # When the main thread exits, daemon threads will
            # continue to run after the threading module is
            # destroyed. Calls to the Event methods can thus
            # result in errors.
            try:
                if not self.__clearEvent.wait(self.__timeout) and \
                   not self.__vetoEvent.is_set():

                    log.debug('Timeout - clearing status')
                    clearStatus()

            except TypeError:
                return



#!/usr/bin/env python
#
# test_progress.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import time
import wx

import numpy as np

from unittest import mock

from . import run_with_wx, simclick, simtext, simkey, realYield

from fsleyes_widgets import isalive
import fsleyes_widgets.utils.progress as progress


def test_Bounce():
    run_with_wx(_test_Bounce)
def _test_Bounce():

    endfuncs = ['Close', 'EndModal', 'Destroy']
    delay    = 200  # msecs
    centis   = delay / 10
    values   = list(np.arange(50))
    passed   = [True]

    for endfunc in endfuncs:

        dlg = progress.Bounce('Title', 'Message', delay=delay, values=values)

        dlg.StartBounce()

        def eval():
            value = dlg.GetValue()

            for i in range(10):

                realYield(centis + 0.5 * centis)
                newval = dlg.GetValue()

                passed[0] = passed[0] and (newval != value)
                value     = newval

            value = dlg.GetValue()
            dlg.StopBounce()
            getattr(dlg, endfunc)()

            realYield((delay * 2) / 10)

            if isalive(dlg):
                passed[0] = passed[0] and (dlg.GetValue() == value)

        if endfunc == 'EndModal':
            wx.CallAfter(eval)
            dlg.ShowModal()
        else:
            dlg.Show()
            eval()
        if endfunc != 'Destroy':
            dlg.Destroy()
        dlg = None
        assert passed[0]


def test_Bounce_manual():
    run_with_wx(_test_Bounce_manual)
def _test_Bounce_manual():

    dlg = progress.Bounce('Title', 'Message', delay=150)

    # make sure it's not bouncing
    val = dlg.GetValue()
    realYield(20)
    assert dlg.GetValue() == val

    # Make sure a bounce changes the value
    dlg.DoBounce()
    assert dlg.GetValue() != val

    # Make sure we can change the message
    val = dlg.GetValue()
    dlg.DoBounce('New message')
    assert dlg.GetValue()   != val
    assert dlg.GetMessage() == 'New message'

    dlg.Destroy()

def test_Bounce_auto_start_stop():
    run_with_wx(_test_Bounce_auto_start_stop)
def _test_Bounce_auto_start_stop():

    dlg = progress.Bounce('Title', 'Message', delay=100)

    dlg.StartBounce()
    val = dlg.GetValue()
    realYield(15)
    assert dlg.GetValue() != val
    val = dlg.GetValue()
    realYield(15)
    assert dlg.GetValue() != val
    val = dlg.GetValue()
    dlg.StopBounce()

    realYield(15)
    assert dlg.GetValue() == val
    realYield(15)
    assert dlg.GetValue() == val

    dlg.StartBounce()
    val = dlg.GetValue()
    realYield(15)
    assert dlg.GetValue() != val
    val = dlg.GetValue()
    realYield(15)
    assert dlg.GetValue() != val
    val = dlg.GetValue()
    dlg.StopBounce()
    dlg.Destroy()


def test_runWithBounce():
    run_with_wx(_test_runWithBounce)
def _test_runWithBounce():


    finished  = [False]
    completed = [False]

    def func():
        for i in range(5):
            time.sleep(0.5)
        finished[0] = True

    def callback(comp):
        completed[0] = comp

    progress.runWithBounce(func,
                           'Title',
                           'Message',
                           delay=100,
                           callback=callback)

    realYield(350)

    assert finished[0] and completed[0]


def test_bounce_ctx():
    run_with_wx(_test_bounce_ctx)
def _test_bounce_ctx():

    with progress.bounce(delay=100) as dlg:
        val = dlg.GetValue()
        realYield(15)
        assert dlg.GetValue() != val
        val = dlg.GetValue()
        realYield(15)
        assert dlg.GetValue() != val


def test_runWithBounce_cancel():
    run_with_wx(_test_runWithBounce_cancel)
def _test_runWithBounce_cancel():

    finished       = [False]
    completed      = [True]
    cancelledcalls = [0]

    def func():
        for i in range(10):
            time.sleep(1)

        finished[0] = True

    def callback(comp):
        completed[0] = comp

    def WasCancelled(s):
        if cancelledcalls[0] > 10:
            return True
        else:
            cancelledcalls[0] += 1
            return False

    dlg = progress.Bounce('Title', 'message', style=wx.PD_CAN_ABORT)

    with mock.patch('wx.ProgressDialog.WasCancelled',
                    WasCancelled):
        progress.runWithBounce(func, dlg=dlg, callback=callback)
        realYield(200)

    dlg.Destroy()

    assert not finished[0] and not completed[0]



#!/usr/bin/env python
#
# test_autotextctrl.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


from . import run_with_wx, simclick, simtext, simkey, realYield, addall

import wx
import fsleyes_widgets.autotextctrl as autott


def sendEvent(target, evType, source=None):

    if source is None:
        source = target

    target.ProcessEvent(wx.CommandEvent(evType, source.GetId()))
    realYield()


# Simple test - programmatically
# set, then retrieve the value
def test_getSet():
    run_with_wx( _test_getSet)
def _test_getSet():

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    atc.ChangeValue('a')
    assert atc.GetValue() == 'a'

    atc.SetValue('b')
    assert atc.GetValue() == 'b'


def test_event():
    run_with_wx( _test_event)
def _test_event():

    called = [None]

    def handler(ev):
        called[0] = ev.text

    sim    = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc    = autott.AutoTextCtrl(parent, modal=False)

    atc.Bind(autott.EVT_ATC_TEXT_ENTER, handler)

    addall(parent, [atc])

    simtext(sim, atc.textCtrl, 'abc')

    assert atc.GetValue() == 'abc'
    assert called[0]      == 'abc'


# Make sure that when the cotrol receives
# focus, its insertion point is at the end
def test_onFocus():
    run_with_wx(_test_onFocus)
def _test_onFocus():

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.ChangeValue('a')
    atc.SetInsertionPoint(0)

    sendEvent(atc, wx.wxEVT_KILL_FOCUS)

    assert atc.GetInsertionPoint() == 0

    sendEvent(atc, wx.wxEVT_SET_FOCUS)
    assert atc.GetInsertionPoint() == 1


# Test showing the popup and selecting a value
def test_popup_select1():
    run_with_wx(_test_popup_select1)
def _test_popup_select1():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    addall(parent, [atc])

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simkey(sim, atc.textCtrl,       wx.WXK_RETURN)
    simkey(sim, atc.popup.textCtrl, wx.WXK_DOWN)
    atc.popup.listBox.SetSelection(1)
    simkey(sim, atc.popup.listBox,  wx.WXK_RETURN)

    assert atc.GetValue() == 'aab'


def test_popup_select2():
    run_with_wx(_test_popup_select2)
def _test_popup_select2():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    addall(parent, [atc])

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])
    simtext(sim, atc.textCtrl, 'b', enter=False)
    simkey( sim, atc.popup.textCtrl, wx.WXK_DOWN)
    simkey( sim, atc.popup.listBox, wx.WXK_RETURN)

    assert atc.GetValue() == 'bcc'


def test_popup_select3():
    run_with_wx(_test_popup_select3)
def _test_popup_select3():

    sim = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    addall(parent, [atc])

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simtext(sim, atc, 'ab', enter=False)

    simkey(sim, atc.popup.textCtrl, wx.WXK_DOWN)
    simkey(sim, atc.popup.listBox,  wx.WXK_RETURN)

    assert atc.GetValue() == 'aba'


def test_popup_dblclick():
    run_with_wx(_test_popup_dblclick)
def _test_popup_dblclick():

    class FakeEv:
        def __init__(self, keycode=None):
            self.keycode = keycode
        def GetKeyCode(self):
            return self.keycode
        def Skip(self):
            pass
        def ResumePropagation(self, a):
            pass

    parent = wx.GetApp().GetTopWindow()
    atc    = autott.AutoTextCtrl(parent, modal=False)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    addall(parent, [atc])

    atc._AutoTextCtrl__onKeyDown(FakeEv(wx.WXK_RETURN))
    atc.popup.listBox.SetSelection(0)
    atc.popup._AutoCompletePopup__onListMouseDblClick(FakeEv())
    realYield()

    assert atc.GetValue() == 'aaa'


def test_popup_cancel():
    run_with_wx(_test_popup_cancel)
def _test_popup_cancel():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    addall(parent, [atc])

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simkey(sim, atc.textCtrl,       wx.WXK_RETURN)
    simkey(sim, atc.popup.textCtrl, wx.WXK_DOWN)
    simkey(sim, atc.popup.listBox,  wx.WXK_DOWN)
    simkey(sim, atc.popup.listBox,  wx.WXK_ESCAPE)

    assert atc.GetValue() == ''


def test_popup_propagate_enter():
    run_with_wx(_test_popup_propagate_enter)
def _test_popup_propagate_enter():

    sim    = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc    = autott.AutoTextCtrl(parent, modal=False)

    called = [False]
    def atcEnter(ev):
        called[0] = True

    addall(parent, [atc])

    atc.Bind(autott.EVT_ATC_TEXT_ENTER, atcEnter)
    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simtext(sim, atc, 'ab', enter=False)

    simkey(sim, atc.popup.textCtrl, wx.WXK_DOWN)
    simkey(sim, atc.popup.listBox,  wx.WXK_RETURN)

    assert atc.GetValue() == 'aba'
    assert called[0]


def test_popup_no_propagate_enter():
    run_with_wx(_test_popup_no_propagate_enter)
def _test_popup_no_propagate_enter():

    sim = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent,
                              modal=False,
                              style=autott.ATC_NO_PROPAGATE_ENTER)

    called = [False]
    def atcEnter(ev):
        called[0] = True


    addall(parent, [atc])

    atc.Bind(autott.EVT_ATC_TEXT_ENTER, atcEnter)
    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simtext(sim, atc, 'ab', enter=False)

    simkey(sim, atc.popup.textCtrl, wx.WXK_DOWN)
    simkey(sim, atc.popup.listBox,  wx.WXK_RETURN)

    assert atc.GetValue() == 'aba'
    assert not called[0]


def test_popup_focusback():
    run_with_wx(_test_popup_focusback)
def _test_popup_focusback():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent, modal=False)

    addall(parent, [atc])

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    simkey(sim,  atc.textCtrl, wx.WXK_RETURN)
    simkey(sim,  atc.popup.textCtrl, wx.WXK_DOWN)
    simkey(sim,  atc.popup.listBox,  wx.WXK_UP)
    simtext(sim, atc.popup.textCtrl, 'abc')

    assert atc.GetValue() == 'abc'



#!/usr/bin/env python
#
# test_floatslider.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import numpy as np

import wx

from unittest import mock
import pytest

from . import run_with_wx, simtext, simclick, simkey, realYield

import fsleyes_widgets.floatslider as floatslider


# Used by test_FloatSlider_logic and test_SliderSpinPanel_logic.
def _test_widget_logic(widget):

    # min, max
    testcases = [
        (     0,      1),
        (    -1,      1),
        (  1e-5,   1e-1),
        (     0,    100),
        (  -100,    100),
        (-2**31,   2**31 - 1),
    ]

    widget.SetMin(0)
    widget.SetMax(1)
    assert np.isclose(widget.GetMin(), 0)
    assert np.isclose(widget.GetMax(), 1)

    # Min > max
    with pytest.raises(ValueError):
        widget.SetRange(1, 0)

    # Min > max
    with pytest.raises(ValueError):
        widget.SetMin(2)

    # Max < min
    with pytest.raises(ValueError):
        widget.SetMax(-1)

    # Min == max
    widget.SetRange(0, 0)
    assert widget.GetRange() == (0, 0)
    assert widget.GetMin() == 0
    assert widget.GetMax() == 0
    widget.SetValue(1)
    assert widget.GetValue() == 0
    widget.SetValue(-1)
    assert widget.GetValue() == 0
    widget.SetValue(0)
    assert widget.GetValue() == 0

    for minv, maxv in testcases:

        widget.SetRange(minv, maxv)
        assert np.all(np.isclose(widget.GetRange(), (minv, maxv)))

        widget.SetValue(minv - 1)
        assert widget.GetValue() == minv
        widget.SetValue(maxv + 1)
        assert widget.GetValue() == maxv

        testvals = np.linspace(minv, maxv, 100)

        for v in testvals:
            widget.SetValue(v)
            assert np.isclose(widget.GetValue(), v)


# Used by test_FloatSlider_logic_integer and
# test_SliderSpinPanel_logic_integer
def _test_widget_logic_integer(widget):

    testcases = [
        (     0,    100),
        (  -100,    100),
        (-2**31,   2**31 - 1),
    ]

    for minv, maxv in testcases:

        widget.SetRange(minv, maxv)

        assert widget.GetRange() == (minv, maxv)

        if minv == -2**31:
            testvals = np.arange(minv, maxv, 2**24)
        else:
            testvals = np.arange(minv, maxv)

        for v in testvals:
            widget.SetValue(v)
            assert widget.GetValue() == v


# You should get it by now
def _test_widget_changeRange(widget):
    widget.SetRange(0, 100)
    widget.SetValue(10)
    widget.SetRange(-100, 100)
    assert np.isclose(widget.GetValue(), 10)
    widget.SetRange(11, 100)
    assert np.isclose(widget.GetValue(), 11)
    widget.SetRange(0, 11)
    assert np.isclose(widget.GetValue(), 11)
    widget.SetRange(0, 5)
    assert np.isclose(widget.GetValue(), 5)


def test_FloatSlider_logic():
    run_with_wx(_test_FloatSlider_logic)
def _test_FloatSlider_logic():
    frame  = wx.GetApp().GetTopWindow()
    slider = floatslider.FloatSlider(frame)
    _test_widget_logic(slider)


def test_FloatSlider_logic_integer():
    run_with_wx(_test_FloatSlider_logic_integer)
def _test_FloatSlider_logic_integer():
    frame  = wx.GetApp().GetTopWindow()
    slider = floatslider.FloatSlider(frame, style=floatslider.FS_INTEGER)
    _test_widget_logic_integer(slider)


def test_FloatSlider_changeRange():
    run_with_wx(_test_FloatSlider_changeRange)
def _test_FloatSlider_changeRange():
    frame  = wx.GetApp().GetTopWindow()
    slider = floatslider.FloatSlider(frame)
    _test_widget_changeRange(slider)


@pytest.mark.dodgy
def test_FloatSlider_mouse_non_gtk():
    run_with_wx(_test_FloatSlider_mouse)

@pytest.mark.dodgy
def test_FloatSlider_mouse_gtk():
    with mock.patch('fsleyes_widgets.floatslider.wx.Platform', '__WXGTK__'):
        run_with_wx(_test_FloatSlider_mouse)

def _test_FloatSlider_mouse():

    sim    = wx.UIActionSimulator()
    frame  = wx.GetApp().GetTopWindow()
    minval = 0
    init   = 50
    maxval = 100
    slider = floatslider.FloatSlider(frame,
                                     value=init,
                                     minValue=minval,
                                     maxValue=maxval)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(slider, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()
    realYield()

    called = [0]

    def handler(ev):
        called[0] += 1

    slider.Bind(wx.EVT_SLIDER, handler)

    xposes = [0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95]

    for xpos in xposes:

        called[0] = 0

        # Make sure clicking on the same spot
        # doesn't result in multiple events
        simclick(sim, slider, pos=[xpos, 0.5])

        expected = minval + xpos * (maxval - minval)

        # Be tolerant with the slider value,
        # because the above xposes may not
        # take into account widget borders.
        tol = (maxval - minval) * 0.05
        assert called[0] == 1
        assert abs(slider.GetValue() - expected) <= tol



def test_FloatSlider_wheel():
    run_with_wx(_test_FloatSlider_wheel)
def _test_FloatSlider_wheel():

    frame  = wx.GetApp().GetTopWindow()
    slider = floatslider.FloatSlider(frame,
                                     value=50,
                                     minValue=0,
                                     maxValue=100,
                                     style=floatslider.FS_MOUSEWHEEL)

    called = [False]

    def handler(ev):
        called[0] = True

    slider.Bind(wx.EVT_SLIDER, handler)

    ev = mock.MagicMock()


    # UIActionSimulator don't do mouse wheel,
    # so we fake a mousewheel event
    val = slider.GetValue()
    ev.GetWheelRotation.return_value = 1
    slider._FloatSlider__onMouseWheel(ev)
    realYield()
    assert called[0]
    assert slider.GetValue() > val


    val = slider.GetValue()

    called[0] = False
    ev.GetWheelRotation.return_value = -1
    slider._FloatSlider__onMouseWheel(ev)
    realYield()
    assert called[0]
    assert slider.GetValue() < val

    val = slider.GetValue()

    called[0] = False
    ev.GetWheelRotation.return_value = 0
    slider._FloatSlider__onMouseWheel(ev)
    realYield()
    assert not called[0]
    assert np.isclose(slider.GetValue(), val)



def test_SliderSpinPanel_logic():
    run_with_wx(_test_SliderSpinPanel_logic)
def _test_SliderSpinPanel_logic():
    frame = wx.GetApp().GetTopWindow()
    panel = floatslider.SliderSpinPanel(frame)
    _test_widget_logic(panel)


def test_SliderSpinPanel_logic_integer():
    run_with_wx(_test_SliderSpinPanel_logic_integer)
def _test_SliderSpinPanel_logic_integer():
    frame  = wx.GetApp().GetTopWindow()
    panel = floatslider.SliderSpinPanel(frame, style=floatslider.SSP_INTEGER)
    _test_widget_logic_integer(panel)


def test_SliderSpinPanel_changeRange():
    run_with_wx(_test_SliderSpinPanel_changeRange)
def _test_SliderSpinPanel_changeRange():
    frame = wx.GetApp().GetTopWindow()
    panel = floatslider.SliderSpinPanel(frame)
    _test_widget_changeRange(panel)


def test_SliderSpinPanel_show_edit_limits():
    run_with_wx(_test_SliderSpinPanel_show_edit_limits)
def _test_SliderSpinPanel_show_edit_limits():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel = floatslider.SliderSpinPanel(
        frame,
        style=floatslider.SSP_SHOW_LIMITS | floatslider.SSP_EDIT_LIMITS)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    panel.SetRange(0, 100)

    result = [None]

    def handler(ev):
        result[0] = (ev.min, ev.max)

    panel.Bind(floatslider.EVT_SSP_LIMIT, handler)

    numberdlg = mock.MagicMock()
    numberdlg.ShowModal.return_value = wx.ID_OK
    numberdlg.GetValue .return_value = 9

    minbtn = panel._SliderSpinPanel__minButton
    maxbtn = panel._SliderSpinPanel__maxButton

    # limbutton, value, expectedrange, shouldTriggerEvent
    testcases = [
        (minbtn,   50, (  50, 100),  True),
        (minbtn,  150, (  50, 100), False),
        (minbtn, -100, (-100, 100),  True),
        (maxbtn,   50, (-100,  50),  True),
        (maxbtn, -200, (-100,  50), False),
        (maxbtn,  500, (-100, 500),  True),

    ]

    with mock.patch('fsleyes_widgets.numberdialog.NumberDialog',
                    return_value=numberdlg):

        for btn, val, expected, shouldEv in testcases:

            result[0] = None

            numberdlg.GetValue.return_value = val

            realYield()
            simclick(sim, btn)
            assert tuple(panel.GetRange()) == expected

            if shouldEv: assert result[0] == expected
            else:        assert result[0] is None


@pytest.mark.dodgy
def test_SliderSpinPanel_events():
    run_with_wx(_test_SliderSpinPanel_events)
def _test_SliderSpinPanel_events():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = floatslider.SliderSpinPanel(frame, label='Value', style=0)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()

    ncalls = [0]
    called = [None]

    def handler(ev):
        ncalls[0] += 1
        called[0] = ev.value

    panel.Bind(floatslider.EVT_SSP_VALUE, handler)

    panel.SetRange(0, 100)

    realYield()
    simclick(sim, panel.slider, pos=[0.5, 0.5])

    assert abs(panel.spinCtrl.GetValue() - 50) < 5
    assert abs(panel.slider  .GetValue() - 50) < 5
    assert abs(panel         .GetValue() - 50) < 5
    assert called[0]                 - 50 < 5
    assert ncalls[0] == 1

    called[0] = None
    ncalls[0] = 0
    simtext(sim, panel.spinCtrl.textCtrl, '75')

    assert np.isclose(panel.spinCtrl.GetValue(), 75)
    assert np.isclose(panel.slider  .GetValue(), 75)
    assert np.isclose(panel         .GetValue(), 75)
    assert np.isclose(called[0],                 75)
    assert ncalls[0] == 1


def test_SliderSpinPanel_nolimit():
    run_with_wx(_test_SliderSpinPanel_nolimit)
def _test_SliderSpinPanel_nolimit():
    frame = wx.GetApp().GetTopWindow()
    panel = floatslider.SliderSpinPanel(frame, style=floatslider.SSP_NO_LIMITS)

    panel.SetRange(0, 100)

    values = [-100, -50, -1, 0, 1, 50, 99, 100, 101, 150, 200]

    for v in values:
        panel.SetValue(v)
        assert panel.GetValue() == v



#!/usr/bin/env python
#
# webpage.py - Convenience functions for opening a URL in a web browser.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides convenience functions for opening a URL in a web
browser.

The following functions are provided:

.. autosummary::
   :nosignatures:

   fileToUrl
   openPage
   openFile
"""

import                   webbrowser
import urllib.parse   as urlparse
import urllib.request as urlrequest


def fileToUrl(fileName):
    """Converts a file path to a URL. """

    return urlparse.urljoin('file:', urlrequest.pathname2url(fileName))


def openPage(url):
    """Opens the given URL in the system-default web browser."""
    webbrowser.open(url)


def openFile(fileName):
    """Opens the given file in the system-default web browser."""
    openPage(fileToUrl(fileName))



#!/usr/bin/env python
#
# test_notebook.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import itertools as it

import pytest

import wx

from . import run_with_wx, simclick, realYield


import fsleyes_widgets.notebook as nb


def nb_run_with_wx(func):
    def wrapper():

        frame = wx.GetApp().GetTopWindow()

        for side, ornt in it.product((wx.HORIZONTAL, wx.VERTICAL),
                                     (wx.LEFT, wx.RIGHT, wx.TOP, wx.BOTTOM)):
            func(side, ornt)
            frame.DestroyChildren()

    run_with_wx(wrapper)


def test_setColours():
    run_with_wx(_test_setColours)
def _test_setColours():
    frame    = wx.GetApp().GetTopWindow()
    notebook = nb.Notebook(frame)

    page1 = wx.Panel(notebook)
    notebook.AddPage(page1, 'page1')

    notebook.SetButtonColours(text='#ffffff',
                              disabledText='#d0d0d0',
                              default='#000000',
                              selected='#0000ff')


def test_add_remove():
    nb_run_with_wx(_test_add_remove)
def _test_add_remove(side, ornt):

    frame    = wx.GetApp().GetTopWindow()
    sizer    = wx.BoxSizer(wx.VERTICAL)
    notebook = nb.Notebook(frame, style=side | ornt)

    sizer.Add(notebook)
    frame.SetSizer(sizer)
    frame.Layout()

    # the notebook should reset the page parent
    page1 = wx.Panel(frame)
    page2 = wx.Panel(notebook)
    page3 = wx.Panel(notebook)

    with pytest.raises(IndexError):
        notebook.InsertPage(-1, page1, 'page1')
    with pytest.raises(IndexError):
        notebook.InsertPage(1,  page1, 'page1')

    assert notebook.PageCount() == 0
    notebook.InsertPage(0, page1, 'page1')
    assert notebook.PageCount() == 1
    notebook.AddPage(page3, 'page3')
    assert notebook.PageCount() == 2
    notebook.InsertPage(1, page2, 'page2')
    assert notebook.PageCount() == 3

    realYield()

    with pytest.raises(IndexError):
        notebook.RemovePage(-1)
    with pytest.raises(IndexError):
        notebook.RemovePage(3)

    assert notebook.pages[0] is page1
    assert notebook.pages[1] is page2
    assert notebook.pages[2] is page3

    assert notebook.FindPage(page1) == 0
    assert notebook.FindPage(page2) == 1
    assert notebook.FindPage(page3) == 2
    notebook.RemovePage(0)
    assert notebook.PageCount() == 2
    assert notebook.FindPage(page1) == wx.NOT_FOUND
    assert notebook.FindPage(page2) == 0
    assert notebook.FindPage(page3) == 1
    notebook.DeletePage(1)
    assert notebook.PageCount() == 1
    assert notebook.FindPage(page1) == wx.NOT_FOUND
    assert notebook.FindPage(page2) == 0
    assert notebook.FindPage(page3) == wx.NOT_FOUND
    notebook.RemovePage(0)
    assert notebook.PageCount() == 0
    assert notebook.FindPage(page1) == wx.NOT_FOUND
    assert notebook.FindPage(page2) == wx.NOT_FOUND
    assert notebook.FindPage(page3) == wx.NOT_FOUND

    realYield()


def test_selection():
    nb_run_with_wx(_test_selection)
def _test_selection(side, ornt):
    frame    = wx.GetApp().GetTopWindow()
    notebook = nb.Notebook(frame, style=side | ornt)
    page1 = wx.Panel(notebook)
    page2 = wx.Panel(notebook)
    page3 = wx.Panel(notebook)

    notebook.AddPage(page1, 'page1')
    notebook.AddPage(page2, 'page2')
    notebook.AddPage(page3, 'page3')

    assert notebook.GetSelection() == 0
    notebook.SetSelection(1)
    assert notebook.GetSelection() == 1
    notebook.SetSelection(2)
    assert notebook.GetSelection() == 2
    notebook.SetSelection(0)
    assert notebook.GetSelection() == 0

    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 1
    notebook.AdvanceSelection(False)
    assert notebook.GetSelection() == 0

    with pytest.raises(IndexError):
        notebook.SetSelection(-1)
    with pytest.raises(IndexError):
        notebook.SetSelection(3)

    notebook.DeletePage(0)
    assert notebook.GetSelection() == 0
    notebook.SetSelection(1)
    notebook.DeletePage(0)
    assert notebook.GetSelection() == 0
    notebook.DeletePage(0)
    assert notebook.GetSelection() is None


def test_enable_disable_show_hide():
    nb_run_with_wx(_test_enable_disable_show_hide)
def _test_enable_disable_show_hide(side, ornt):
    frame    = wx.GetApp().GetTopWindow()
    sizer    = wx.BoxSizer(wx.VERTICAL)
    notebook = nb.Notebook(frame, style=side | ornt)
    page1    = wx.Panel(notebook)
    page2    = wx.Panel(notebook)
    page3    = wx.Panel(notebook)

    notebook.AddPage(page1, 'page1')
    notebook.AddPage(page2, 'page2')
    notebook.AddPage(page3, 'page3')

    assert notebook.GetSelection() == 0
    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 1
    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 2
    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 0
    notebook.AdvanceSelection(False)
    assert notebook.GetSelection() == 2
    notebook.SetSelection(0)
    notebook.DisablePage(1)
    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 2
    notebook.AdvanceSelection(False)
    assert notebook.GetSelection() == 0
    notebook.EnablePage(1)

    notebook.SetSelection(1)
    notebook.HidePage(1)
    assert notebook.GetSelection() == 2
    notebook.AdvanceSelection(False)
    assert notebook.GetSelection() == 0
    notebook.ShowPage(1)
    notebook.AdvanceSelection()
    assert notebook.GetSelection() == 1


@pytest.mark.dodgy
def test_notebook_events():
    nb_run_with_wx(_test_notebook_events)
def _test_notebook_events(side, ornt):
    sim      = wx.UIActionSimulator()
    frame    = wx.GetApp().GetTopWindow()
    sizer    = wx.BoxSizer(wx.VERTICAL)
    notebook = nb.Notebook(frame, style=side | ornt)
    page1    = wx.Panel(notebook)
    page2    = wx.Panel(notebook)
    page3    = wx.Panel(notebook)

    notebook.AddPage(page1, 'page1')
    notebook.AddPage(page2, 'page2')
    notebook.AddPage(page3, 'page3')

    clicked = [None]

    def onbtn(ev):
        clicked[0] = ev.index

    notebook.Bind(nb.EVT_PAGE_CHANGE, onbtn)

    sizer.Add(notebook, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()
    realYield()

    btn1, btn2, btn3 = notebook.buttons

    notebook.SetSelection(2)

    simclick(sim, btn1)
    assert clicked[0] == 0
    simclick(sim, btn2)
    assert clicked[0] == 1
    simclick(sim, btn3)
    assert clicked[0] == 2



#!/usr/bin/env python
#
# textbitmap.py - A function which renders some text using matplotlib, and
# returns it as an RGBA bitmap.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a single function, :func:`textBitmap`, which renders
some text off-screen using :mod:`matplotlib`, and returns it as an RGBA bitmap.
"""


POINT_SIZE = 1 / 72
"""Size of one point in inches at 72 dpi. Font sizes are specified in points at
72 dpi - this value is used to convert from font size to inches (and on to
pixels).
"""


def textBitmap(text,
               width=None,
               height=None,
               fontSize=None,
               fgColour=None,
               bgColour=None,
               alpha=1.0,
               fontFamily=None,
               halign=None,
               dpi=96):
    """Draw some text using :mod:`matplotlib`.

    The rendered text is returned as a RGBA bitmap within a ``numpy.uint8``
    array of size :math:`h \\times w \\times 4`, with the top-left pixel
    located at index ``[0, 0, :]``.

    At least one of the ``fontSize`` or the ``height`` arguments need to be
    specified - if one of these is provided, the other size options will be
    inferred, although the inference procedure does not support multi-line
    text.

    :arg text:       Text to render.

    :arg width:      Width in pixels.

    :arg height:     Height in pixels.

    :arg fontSize:   Font size in points.

    :arg fgColour:   Foreground (text) colour - can be any colour specification
                     that is accepted by :mod:`matplotlib`.

    :arg bgColour:   Background colour  - can be any colour specification that
                     is accepted by :mod:`matplotlib`..

    :arg alpha:      Text transparency, in the range ``[0.0 - 1.0]``.

    :arg fontFamily: Font family, e.g. ``'monospace'`` or ``'sans-serif'``,
                     defaults to matplotlib default.

    :arg dpi:        Dots per inch, defaults to 96.

    :arg halign:     Horizontal alignment - one of ``'centre'`` (default),
                     ``'center'``, ``'left'`` or ``right'``.

    :returns:        ``numpy.uint8`` array of size
                     :math:`h \\times w \\times 4`
    """

    if text is None or text.strip() == '':
        raise ValueError('Some text must be specified.')

    if (fontSize is None) and (height is None):
        raise ValueError('One of fontSize or height must be specified.')

    if halign in (None, 'centre'):
        halign = 'center'

    # Imports are expensive
    import numpy                           as np
    import matplotlib.backends.backend_agg as mplagg
    import matplotlib.transforms           as mplxf
    import matplotlib.figure               as mplfig

    # convert points to pixels or vice versa.
    # Estimate width from font size if not
    # provided - we will crop the result
    # afterwards.
    crop   = width is None
    nlines = text.count('\n') + 1
    if fontSize is None: fontSize =          height   / POINT_SIZE / dpi
    if height   is None: height   = nlines * fontSize * POINT_SIZE * dpi
    if width    is None: width    = max(fontSize, fontSize * len(text))
    if fgColour is None: fgColour = '#000000'

    fig    = mplfig.Figure(figsize=(width / dpi, height / dpi), dpi=dpi)
    canvas = mplagg.FigureCanvasAgg(fig)
    ax     = fig.add_axes([0, 0, 1, 1])
    ax.axis('off')

    if bgColour is not None: fig.patch.set_facecolor(bgColour)
    else:                    fig.patch.set_alpha(0)

    ax.set_xticks([])
    ax.set_yticks([])

    if   halign == 'left':  tx = 0.0
    elif halign == 'right': tx = 1.0
    else:                   tx = 0.5

    textobj = ax.text(tx,
                      0.5,
                      text,
                      fontsize=fontSize,
                      verticalalignment='center',
                      horizontalalignment=halign,
                      transform=ax.transAxes,
                      color=fgColour,
                      alpha=alpha,
                      fontfamily=fontFamily)

    # if a width wasn't specified, we auto-
    # fit the bitmap to the rendered text
    if crop:
        # tight bounding box around text
        bbox = textobj.get_window_extent(renderer=canvas.get_renderer())

        # a tiny amount seems to get cropped on the right, for
        # centre/right aligned text. So we shift the text left
        # a little, and add some padding to the figure size.
        if halign == 'left': offset = 0
        else:                offset = POINT_SIZE * 2

        textobj.set_transform(mplxf.offset_copy(
            textobj.get_transform(), fig, -offset, 0))

        fig.set_size_inches((offset + bbox.width / dpi, height / dpi))

    canvas.draw()

    # get the bitmap data and reshape
    # and transpose it to (H, W, [RGBA])
    ncols, nrows = canvas.get_width_height()
    bitmap       = canvas.tostring_argb()
    bitmap       = np.frombuffer(bitmap, dtype=np.uint8)
    bitmap       = bitmap.reshape(nrows, ncols, 4)
    rgb          = bitmap[:, :, 1:]
    a            = bitmap[:, :, 0]
    bitmap       = np.dstack((rgb, a))

    return bitmap



#!/usr/bin/env python
#
# widgetgrid.py - A tabular grid of widgets.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WidgetGrid` class, which can display a
tabular grid of arbitrary widgets.
"""


import functools as ft
import              logging

import wx
import wx.lib.newevent as wxevent

import fsleyes_widgets.utils.b64icon as b64icon


log = logging.getLogger(__name__)


class WidgetGrid(wx.ScrolledWindow):
    """A scrollable panel which displays a tabular grid of widgets.  A
    ``WidgetGrid`` looks something like this:

      .. image:: images/widgetgrid.png
        :scale: 50%
        :align: center


    The most important methods are:

     .. autosummary::
        :nosignatures:

        GetGridSize
        SetGridSize
        Refresh
        DeleteRow
        InsertRow
        SetColours
        SetWidget
        SetText
        ClearGrid


    *Labels*

    .. autosummary::
       :nosignatures:

       ShowRowLabels
       ShowColLabels
       SetRowLabel
       SetColLabel


    *Selections*

    .. autosummary::
       :nosignatures:

       GetSelection
       SetSelection


    *Styles*


    The ``WidgetGrid`` supports the following styles:

    =============================== ==================================
    ``wx.HSCROLL``                  Use a horizontal scrollbar.
    ``wx.VSCROLL``                  Use a vertical scrollbar.
    :data:`WG_SELECTABLE_CELLS`     Individual cells are selectable.
    :data:`WG_SELECTABLE_ROWS`      Rows are selectable.
    :data:`WG_SELECTABLE_COLUMN`    Columns are selectable.
    :data:`WG_KEY_NAVIGATION`       The keyboard can be used for
                                    navigation.
    :data:`WG_DRAGGABLE_COLUMNS`    Columns can be dragged to re-order
                                    them (see also the
                                    :meth:`ReorderColumns` method)
    =============================== ==================================


    The ``*_SELECTABLE_*`` styles are mutualliy exclusive; their precedence
    is equivalent to their order in the above table. By default, the arrow
    keys are used for keyboard navigation, but these are customisable via
    the :meth:`SetNavKeys` method.


    *Events*

    The following events may be emitted by a ``WidgetGrid``:

    .. autosummary::
       :nosignatures:

       :data:`WidgetGridSelectEvent`
       :data:`WidgetGridReorderEvent`
    """


    _defaultBorderColour = None
    """The colour of border a border which is shown around every cell in the
    grid. Initialised in :meth:`__init__`.
    """


    _defaultOddColour    = None
    """Background colour for cells on odd rows.
    Initialised in :meth:`__init__`.
    """


    _defaultEvenColour   = None
    """Background colour for cells on even rows.
    Initialised in :meth:`__init__`.
    """


    _defaultLabelColour  = None
    """Background colour for row and column labels.
    Initialised in :meth:`__init__`.
    """


    _defaultSelectedColour = None
    """Background colour for selected cells.
    Initialised in :meth:`__init__`.
    """


    _defaultDragColour = None
    """Background colour for columns being dragged.
    Initialised in :meth:`__init__`.
    """


    def __init__(self, parent, style=None):
        """Create a ``WidgetGrid``.

        :arg parent: The :mod:`wx` parent object.
        :arg style:  Style flags  - can be a combination of ``wx.HSCROLL``,
                     ``wx.VSCROLL``, :data:`WG_SELECTABLE_CELLS`,
                     :data:`WG_SELECTABLE_ROWS`,
                     :data:`WG_SELECTABLE_COLUMNS`, :data:`WG_KEY_NAVIGATION`,
                      and :data:`WG_DRAGGABLE_COLUMNS`.
        """

        border = wx.SystemSettings.GetColour(wx.SYS_COLOUR_ACTIVEBORDER)
        odd    = wx.SystemSettings.GetColour(wx.SYS_COLOUR_LISTBOX)
        even   = wx.SystemSettings.GetColour(wx.SYS_COLOUR_LISTBOX)\
                                  .ChangeLightness(90)
        label  = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
        select = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        drag   = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)

        if WidgetGrid._defaultBorderColour is None:
            WidgetGrid._defaultBorderColour = border
        if WidgetGrid._defaultOddColour is None:
            WidgetGrid._defaultOddColour = odd
        if WidgetGrid._defaultEvenColour is None:
            WidgetGrid._defaultEvenColour = even
        if WidgetGrid._defaultLabelColour is None:
            WidgetGrid._defaultLabelColour = label
        if WidgetGrid._defaultSelectedColour is None:
            WidgetGrid._defaultSelectedColour = select
        if WidgetGrid._defaultDragColour is None:
            WidgetGrid._defaultDragColour = drag

        if style is None:
            style = wx.HSCROLL | wx.VSCROLL

        self.__hscroll        = style & wx.HSCROLL
        self.__vscroll        = style & wx.VSCROLL
        self.__keynav         = style & WG_KEY_NAVIGATION
        self.__draggable      = style & WG_DRAGGABLE_COLUMNS
        self.__dragLimit      = -1
        self.__dragStartCol   = None
        self.__dragCurrentCol = None

        if   style & WG_SELECTABLE_CELLS:   self.__selectable = 'cells'
        elif style & WG_SELECTABLE_ROWS:    self.__selectable = 'rows'
        elif style & WG_SELECTABLE_COLUMNS: self.__selectable = 'columns'
        else:
            self.__keynav     = False
            self.__selectable = None

        # clear WG_* flags before passing
        # style to ScrolledWindow
        style = (style & wx.HSCROLL) | (style & wx.VSCROLL) | wx.WANTS_CHARS
        wx.ScrolledWindow.__init__(self, parent, style=style)

        hrate = 1 if self.__hscroll else 0
        vrate = 1 if self.__vscroll else 0
        self.SetScrollRate(hrate, vrate)

        self.__gridPanel = wx.Panel(self, style=wx.WANTS_CHARS)

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.__sizer)

        # if column drag is enabled, we use a separate
        # strip above the grid to show where a dragged
        # column will be placed when it is dropped.
        if self.__draggable:

            self.__dragIcon  = b64icon.loadBitmap(TRIANGLE_ICON)
            height           = self.__dragIcon.GetSize()[1]
            self.__dragPanel = wx.Window(self)
            self.__dragPanel.SetMinSize((-1, height))
            self.__dragPanel.SetMaxSize((-1, height))
            self.__sizer.Add(self.__dragPanel, flag=wx.EXPAND)

            self.__dragPanel.Bind(wx.EVT_PAINT, self.__dragPanelPaint)

        self.__sizer.Add(self.__gridPanel, flag=wx.EXPAND)

        # The __widgets array contains wx.Panel
        # objects which are used as containers
        # for all widgets added to the grid.
        # The __widgetRefs array contains the
        # actual widget objects that were passed
        # to the SetWidget method.
        self.__gridSizer      = None
        self.__nrows          = 0
        self.__ncols          = 0
        self.__widgets        = []
        self.__widgetRefs     = []
        self.__rowLabels      = []
        self.__colLabels      = []
        self.__selected       = None

        self.__showRowLabels  = False
        self.__showColLabels  = False
        self.__borderColour   = WidgetGrid._defaultBorderColour
        self.__labelColour    = WidgetGrid._defaultLabelColour
        self.__oddColour      = WidgetGrid._defaultOddColour
        self.__evenColour     = WidgetGrid._defaultEvenColour
        self.__selectedColour = WidgetGrid._defaultSelectedColour
        self.__dragColour     = WidgetGrid._defaultDragColour
        self.__upKey          = wx.WXK_UP
        self.__downKey        = wx.WXK_DOWN
        self.__leftKey        = wx.WXK_LEFT
        self.__rightKey       = wx.WXK_RIGHT

        self.Bind(wx.EVT_SIZE, self.__onResize)

        if self.__keynav:
            # We use CHAR_HOOK for key events,
            # because we want to capture key
            # presses whenever this panel or
            # any of its children has focus.
            self.Bind(wx.EVT_CHAR_HOOK, self.__onKeyboard)

        if self.__selectable:
            # A silly internal multi-level semaphore
            # used to ignore child focus events when
            # this WidgetGrid generated them.  See
            # the __selectCell method for more silly
            # comments.
            self.__ignoreFocus = 0
            self.Bind(wx.EVT_CHILD_FOCUS, self.__onChildFocus)


    @property
    def rowLabels(self):
        """Returns the ``wx.StaticText`` objects used for the row labels. """
        return [l[1] for l in self.__rowLabels]


    @property
    def colLabels(self):
        """Returns the ``wx.StaticText`` objects used for the column labels.
        """
        return [l[1] for l in self.__colLabels]


    @property
    def widgets(self):
        """Returns a list of lists, containing all widgets in the grid.
        """
        return [list(row) for row in self.__widgets]


    def Refresh(self):
        """Redraws the contents of this ``WidgetGrid``. This method must be
        called after the contents of the grid are changed.
        """
        self.__refresh()


    def __recurse(self, obj, funcname, *args, **kwargs):
        """Recursively call ``funcname`` on ``obj`` and all its children.

        This really is something which ``wxwidgets`` should be able to do for
        me (e.g. enable/disable a window *and* all of its children).
        """
        if obj is self:
            func = ft.partial(getattr(wx.ScrolledWindow, funcname), self)
        else:
            func = getattr(obj, funcname)

        func(*args, **kwargs)

        for child in obj.GetChildren():
            self.__recurse(child, funcname, *args, **kwargs)


    def Disable(self):
        """Disables this ``WidgetGrid``. """
        self.Enable(False)


    def Enable(self, enable=True):
        """Enables/disable this ``WidgetGrid``, and recursively does the same
        to all of its children.
        """
        self.__recurse(self, 'Enable', enable)


    def Hide(self):
        """Hides this ``WidgetGrid``. """
        self.Show(False)


    def Show(self, show=True):
        """Shows/hides this ``WidgetGrid``, and recursively does the same
        to all of its children.
        """
        self.__recurse(self, 'Show', show)


    def SetEvtHandlerEnabled(self, enable=True):
        """Enables/disables events on this ``WidgetGrid``, and recursively does
        the same to all of its children.
        """
        self.__recurse(self, 'SetEvtHandlerEnabled', enable)


    def SetColours(self, **kwargs):
        """Set the colours used in this ``WidgetGrid``. The :meth:`Refresh`
        method must be called afterwards for this method to take effect.

        :arg border:   The cell border colour.

        :arg label:    Background colour for row and column labels.

        :arg odd:      Background colour for cells on odd rows.

        :arg even:     Background colour for cells on even rows.

        :arg selected: Background colour for selected cells.

        :arg drag:     Background colour for columns being dragged.
        """

        border   = kwargs.get('border',   self)
        label    = kwargs.get('label',    self)
        odd      = kwargs.get('odd',      self)
        even     = kwargs.get('even',     self)
        selected = kwargs.get('selected', self)
        drag     = kwargs.get('drag',     self)

        if border   is not self: self.__borderColour   = border
        if label    is not self: self.__labelColour    = label
        if odd      is not self: self.__oddColour      = odd
        if even     is not self: self.__evenColour     = even
        if selected is not self: self.__selectedColour = selected
        if drag     is not self: self.__dragColour     = drag


    def SetNavKeys(self, **kwargs):
        """Set the keys used for keyboard navigation (if the
        :data:`WG_KEY_NAVIGATION` style is enabled). Setting an argument
        to ``None`` will disable navigation in that direction.

        :arg up:    Key to use for up navigation.
        :arg down:  Key to use for down navigation.
        :arg left:  Key to use for left navigation.
        :arg right: Key to use for right navigation.
        """

        up    = kwargs.get('up',    self)
        down  = kwargs.get('down',  self)
        left  = kwargs.get('left',  self)
        right = kwargs.get('right', self)

        self.__upKey    = up
        self.__downKey  = down
        self.__leftKey  = left
        self.__rightKey = right


    def SetDragLimit(self, limit):
        """Set the index of the highest column that can be dragged.  Only
        columns before this limit can be dragged, and they can only be dropped
        onto a location before the limit. Only relevant if
        :data:`WG_DRAGGABLE_COLUMNS` is enabled.
        """
        self.__dragLimit = limit


    def __onResize(self, ev):
        """Called when this ``WidgetGrid`` is resized. Makes sure the
        scrollbars are up to date.
        """
        self.FitInside()


    def __reparent(self, widget, parent):
        """Convenience method which  re-parents the given widget. If
        ``widget`` is a :class:`wx.Sizer` the sizer children are re-parented.
        """
        if isinstance(widget, wx.Sizer):
            widget = [c.GetWindow() for c in widget.GetChildren()]
        else:
            widget = [widget]

        for w in widget:
            w.Reparent(parent)


    def __setBackgroundColour(self, widget, colour):
        """Convenience method which changes the background colour of the given
        widget. If ``widget`` is a :class:`wx.Sizer` the background colours of
        the sizer children is updated.
        """
        if isinstance(widget, wx.Sizer):
            widget = [c.GetWindow() for c in widget.GetChildren()]
        else:
            widget = [widget]

        for w in widget:
            w.SetBackgroundColour(colour)


    def __refresh(self):
        """Lays out and re-sizes the entire widget grid. """

        borderColour = self.__borderColour
        labelColour  = self.__labelColour
        oddColour    = self.__oddColour
        evenColour   = self.__evenColour
        ncols        = self.__ncols

        if borderColour is None: borderColour = WidgetGrid._defaultBorderColour
        if labelColour  is None: labelColour  = WidgetGrid._defaultLabelColour
        if oddColour    is None: oddColour    = WidgetGrid._defaultOddColour
        if evenColour   is None: evenColour   = WidgetGrid._defaultEvenColour

        # Grid is empty
        if self.__gridSizer is None:
            self.FitInside()
            self.Layout()
            return

        self.__gridPanel.SetBackgroundColour(borderColour)

        # Clear the sizer per-item, as the
        # wx.Sizer.Clear will destroy any
        # child sizers, and we don't want that
        for i in reversed(range(self.__gridSizer.GetItemCount())):
            self.__gridSizer.Detach(i)

        # empty cell in top left of grid
        self.__gridSizer.Add((-1, -1), flag=wx.EXPAND)

        # column labels
        for coli, (lblPanel, colLabel) in enumerate(self.__colLabels):

            lblPanel._wg_row = -1
            lblPanel._wg_col = coli
            colLabel._wg_row = -1
            colLabel._wg_col = coli

            # If drag limit is set, add a border
            # between the last draggable column,
            # unless all columns are draggable.
            if (coli == self.__dragLimit) and (coli < ncols - 1):
                flag   = wx.EXPAND | wx.RIGHT
                border = 2
            else:
                flag   = wx.EXPAND
                border = 0

            lblPanel.SetBackgroundColour(labelColour)
            colLabel.SetBackgroundColour(labelColour)

            self.__gridSizer.Add( lblPanel, border=border, flag=flag)
            self.__gridSizer.Show(lblPanel, self.__showColLabels)

        # Rows
        for rowi, (lblPanel, rowLabel) in enumerate(self.__rowLabels):

            lblPanel._wg_row = rowi
            lblPanel._wg_col = -1
            rowLabel._wg_row = rowi
            rowLabel._wg_col = -1

            if rowi == self.__nrows - 1: flag = wx.TOP | wx.LEFT | wx.BOTTOM
            else:                        flag = wx.TOP | wx.LEFT

            lblPanel.SetBackgroundColour(labelColour)
            rowLabel.SetBackgroundColour(labelColour)

            self.__gridSizer.Add( lblPanel, flag=wx.EXPAND)
            self.__gridSizer.Show(lblPanel, self.__showRowLabels)

            # Widgets
            for coli in range(self.__ncols):

                widget    = self.__widgetRefs[rowi][coli]
                container = self.__widgets[   rowi][coli]

                widget   ._wg_row = rowi
                widget   ._wg_col = coli
                container._wg_row = rowi
                container._wg_col = coli

                # border at drag limit
                if (coli == self.__dragLimit) and (coli < ncols - 1):
                    flag   = wx.EXPAND | wx.RIGHT
                    border = 2
                else:
                    flag   = wx.EXPAND
                    border = 0

                self.__gridSizer.Add(container,
                                     flag=flag,
                                     border=border,
                                     proportion=1)

                if rowi % 2: colour = oddColour
                else:        colour = evenColour

                self.__setBackgroundColour(container, colour)
                self.__setBackgroundColour(widget,    colour)

        if self.__selected is not None:
            row, col = self.__selected
            self.SetSelection(row, col)

        self.FitInside()
        self.Layout()


    def GetGridSize(self):
        """Returns the current grid size as a tuple containing ``(rows, cols)``.
        """
        return self.__nrows, self.__ncols


    def SetGridSize(self, nrows, ncols, growCols=None):
        """Set the size of the widdget grid. The :meth:`Refresh` method must
        be called afterwards for this method to take effect.

        :arg nrows:    Number of rows

        :arg ncols:    Number of columns

        :arg growCols: A sequence specifying which columns should be stretched
                       to fit.
        """

        if nrows < 0 or ncols < 0:
            raise ValueError('Invalid size ({}, {})'.format(nrows, ncols))

        # If the caller has not specified which columns
        # should stretch, then stretch them all so the
        # grid is sized to fit the available space.
        if growCols is None:
            growCols = range(ncols)

        self.ClearGrid()

        self.__nrows     = nrows
        self.__ncols     = ncols
        self.__dragLimit = -1

        # set hgap and vgap so we get
        # a 1px border between cells
        self.__gridSizer = wx.FlexGridSizer(nrows + 1, ncols + 1, 1, 1)
        self.__gridSizer.SetFlexibleDirection(wx.BOTH)

        for col in growCols:
            self.__gridSizer.AddGrowableCol(col + 1)

        self.__widgets    = [[None] * ncols for i in range(nrows)]
        self.__widgetRefs = [[None] * ncols for i in range(nrows)]
        self.__rowLabels  =  [None] * nrows
        self.__colLabels  =  [None] * ncols

        for rowi in range(nrows):
            for coli in range(ncols):
                self.__initCell(rowi, coli)

        for rowi in range(nrows): self.__initRowLabel(rowi)
        for coli in range(ncols): self.__initColLabel(coli)

        # Put the main grid sizer inside
        # another sizer, so we get a
        # border around the entire thing
        # (this can't be done in the grid
        # sizer itself, because the borders
        # between columns potentially vary)
        self.__gridBorderSizer = wx.BoxSizer(wx.VERTICAL)
        self.__gridBorderSizer.Add(self.__gridSizer,
                                   flag=wx.EXPAND | wx.ALL,
                                   border=1)

        self.__gridPanel.SetSizer(self.__gridBorderSizer)


    def __initCell(self, row, col):
        """Called by :meth:`SetGridSize` and :meth:`InsertRow`. Creates a
        placeholder ``wx.Panel`` at the specified cell.
        """
        placeholder = wx.Panel(self.__gridPanel)
        self.__widgets[   row][col] = placeholder
        self.__widgetRefs[row][col] = placeholder


    def __initRowLabel(self, row):
        """Called by :meth:`SetGridSize` and :meth:`InsertRow`. Creates a
        label widget at the specified row.
        """
        panel = wx.Panel(self.__gridPanel)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        lbl   = wx.StaticText(
            panel,
            style=wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)

        # See comment in SetWidget
        panel._wg_cell = True

        panel.SetSizer(sizer)
        sizer.Add(lbl, flag=wx.CENTRE)

        self.__initWidget(panel, row, -1)
        self.__initWidget(lbl,   row, -1)
        self.__rowLabels[row] = (panel, lbl)


    def __initColLabel(self, col):
        """Called by :meth:`SetGridSize`. Creates a label widget at the
        specified column
        """
        panel = wx.Panel(self.__gridPanel)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        label = wx.StaticText(
            panel,
            style=wx.ALIGN_CENTRE_HORIZONTAL | wx.ALIGN_CENTRE_VERTICAL)

        # See comment in SetWidget
        panel._wg_cell = True

        panel.SetSizer(sizer)
        sizer.Add(label, flag=wx.CENTRE)

        self.__initWidget(panel, -1, col)
        self.__initWidget(label, -1, col)
        self.__colLabels[col] = (panel, label)

        if self.__draggable:
            label.Bind(wx.EVT_LEFT_DOWN, self.__onColumnLabelMouseDown)
            label.Bind(wx.EVT_LEFT_UP,   self.__onColumnLabelMouseUp)
            label.Bind(wx.EVT_MOTION,    self.__onColumnLabelMouseDrag)
            panel.Bind(wx.EVT_LEFT_DOWN, self.__onColumnLabelMouseDown)
            panel.Bind(wx.EVT_LEFT_UP,   self.__onColumnLabelMouseUp)
            panel.Bind(wx.EVT_MOTION,    self.__onColumnLabelMouseDrag)


    def __getCellPanel(self, widget):
        """Returns the parent ``wx.Panel`` for the given ``widget``, or
        ``None`` if the widget is not in the grid.
        """

        while widget is not None:

            if hasattr(widget, '_wg_cell'):
                break

            widget = widget.GetParent()

        return widget


    def GetRow(self, widget):
        """Returns the index of the row in which the given ``widget`` is
        located, or ``-1`` if it is not in the ``WidgetGrid``.
        """
        try:                   return widget._wg_row
        except AttributeError: return -1


    def GetColumn(self, widget):
        """Returns the index of the column in which the given ``widget`` is
        located, or ``-1`` if it is not in the ``WidgetGrid``.
        """
        try:                   return widget._wg_col
        except AttributeError: return -1


    def DeleteRow(self, row):
        """Removes the specified ``row`` from the grid, destroying all
        widgets on that row. This method does not need to be followed
        by a call to :meth:`Refresh`, but a call to ``Layout`` may be
        required.

        .. note:: Make sure you reparent any widgets that you do not want
                  destroyed before calling this method.
        """

        if self.__gridSizer is None:
            raise ValueError('No grid')

        if row < 0 or row >= self.__nrows:
            raise ValueError('Invalid row index {}'.format(row))

        log.debug('Deleting row {} (sizer indices {} - {})'.format(
            row,
            (row + 1) * (self.__ncols + 1),
            (row + 1) * (self.__ncols + 1) + self.__ncols))

        # Remove from the grid
        for col in reversed(range(self.__ncols + 1)):
            self.__gridSizer.Detach((row + 1) * (self.__ncols + 1) + col)

        # Update row/col references
        # on all widgets/labels
        for ri in range(row, self.__nrows):
            for ci in range(self.__ncols):
                self.__widgetRefs[ri][ci]._wg_row -= 1
                self.__widgets[   ri][ci]._wg_row -= 1

            self.__rowLabels[ri][0]._wg_row -= 1
            self.__rowLabels[ri][1]._wg_row -= 1

        # Destroy the widgets and the row label
        for widget in self.__widgets[row]:
            widget.Destroy()

        self.__rowLabels .pop(row)[0].Destroy()

        # Remove references to them
        self.__widgetRefs.pop(row)
        self.__widgets   .pop(row)

        # Update the grid
        self.__nrows -= 1
        self.__gridSizer.SetRows(self.__nrows + 1)

        # Update selected widget if necessary
        if self.__selected is not None:
            srow, scol = self.__selected
            if   srow == row:             self.__selected = None
            elif srow > row and srow > 0: self.__selected = (srow - 1, scol)

        self.FitInside()


    def InsertRow(self, row):
        """Inserts a new row into the ``WidgetGrid`` at the specified ``row``
        index. This method must be followed by a call to :meth:`Refresh`.
        """

        if self.__gridSizer is None:
            raise ValueError('No grid')

        if row < 0:
            raise ValueError('Invalid row index {}'.format(row))

        if row >= self.__nrows:
            row = self.__nrows

        log.debug('Inserting row at {}'.format(row))

        # Add empty label/cell
        # values for the new row
        self.__rowLabels .insert(row,  None)
        self.__widgets   .insert(row, [None] * self.__ncols)
        self.__widgetRefs.insert(row, [None] * self.__ncols)

        # Update the grid
        self.__nrows += 1
        self.__gridSizer.SetRows(self.__nrows + 1)

        # Initialise the contents
        # of the new row
        self.__initRowLabel(row)
        for col in range(self.__ncols):
            self.__initCell(row, col)

        # update row/col indices
        for ri in range(row + 1, self.__nrows):
            for ci in range(self.__ncols):
                self.__widgetRefs[ri][ci]._wg_row += 1
                self.__widgets[   ri][ci]._wg_row += 1

            self.__rowLabels[ri][0]._wg_row += 1
            self.__rowLabels[ri][1]._wg_row += 1

        # Update selected widget if necessary
        if self.__selected is not None:
            srow, scol = self.__selected

            if srow >= row and srow < self.__nrows - 1:
                self.__selected = (srow + 1, scol)


    def ClearGrid(self):
        """Removes and destroys all widgets from the grid, and sets the grid
        size to ``(0, 0)``. The :meth:`Refresh` method must be called
        afterwards for this method to take effect.
        """

        if self.__gridSizer is not None:
            self.__gridSizer.Clear(True)

        self.__gridSizer  = None
        self.__nrows      = 0
        self.__ncols      = 0
        self.__dragLimit  = -1
        self.__widgets    = []
        self.__widgetRefs = []
        self.__rowLabels  = []
        self.__colLabels  = []
        self.__selected   = None

        self.__gridPanel.SetSizer(None)


    def SetText(self, row, col, text):
        """Convenience method which creates a :class:`wx.StaticText` widget
        with the given text, and passes it to the :meth:`SetWidget` method.

        If there is already a ``wx.StaticText`` widget at the given
        ``row``/``col``, it is re-used, and its label simply updated.

        :arg row:  Row index.

        :arg col:  Column index.

        :arg text: Text to display.
        """

        txt = self.GetWidget(row, col)

        if isinstance(txt, wx.StaticText):
            txt.SetLabel(text)
        else:
            txt = wx.StaticText(self.__gridPanel,
                                label=text,
                                style=wx.WANTS_CHARS)
            self.SetWidget(row, col, txt)


    def GetWidget(self, row, col):
        """Returns the widget located at the specified row/column. """

        return self.__widgetRefs[row][col]


    def SetWidget(self, row, col, widget):
        """Adds the given widget to the grid. The :meth:`Refresh` method
        must be called afterwards for this method to take effect.

        The parent of the widget is changed to this ``WidgetGrid``.

         .. note:: The provided widget may alternately be a
                   :class:`wx.Sizer`. However, nested sizers, i.e. sizers
                   which contain other sizers, are not supported.

        :arg row:    Row index.

        :arg col:    Column index.

        :arg widget: The widget or sizer to add to the grid.

        Raises an :exc:`IndexError` if the specified grid location ``(row,
        col)`` is invalid.
        """
        if row <  0            or \
           col <  0            or \
           row >= self.__nrows or \
           col >= self.__ncols:
            raise IndexError('Grid location ({}, {}) out of bounds ({}, {})'.
                             format(row, col, self.__nrows, self.__ncols))

        # Embed the widget in a panel,
        # as Linux/GTK has trouble
        # changing the background colour
        # of some controls
        panel = wx.Panel(self.__gridPanel, style=wx.WANTS_CHARS)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sizer)

        # Put a marker on the cell panel
        # so the __gelCellPanel method
        # can identify it - cell panels
        # are used in certain event
        # handlers (e.g. __onColumnLabelMouse*)
        panel._wg_cell = True

        self.__reparent(widget, panel)

        self.__initWidget(widget, row, col)
        self.__initWidget(panel,  row, col)

        sizer.Add(widget, flag=wx.EXPAND, proportion=1)

        if self.__widgets[row][col] is not None:
            self.__widgets[row][col].Destroy()

        self.__widgetRefs[row][col] = widget
        self.__widgets[   row][col] = panel


    def __initWidget(self, widget, row, col):
        """Called by the :meth:`AddWidget` method.

        Performs some initialisation on a widget which has just been added to
        the grid.

        :arg widget: The widget to initialise
        :arg row:    Row index of the widget in the grid.
        :arg col:    Column index of the widget in the grid.
        """

        def scroll(ev):
            posx, posy = self.GetViewStart()
            rotation   = ev.GetWheelRotation()

            if   rotation > 0: delta =  5
            elif rotation < 0: delta = -5
            else:              return

            if ev.GetWheelAxis() == wx.MOUSE_WHEEL_VERTICAL: posy -= delta
            else:                                            posx += delta

            self.Scroll(posx, posy)

        def initWidget(w):

            # Under Linux/GTK, we need to bind a mousewheel
            # listener to every child of the panel in order
            # for scrolling to work correctly. This is not
            # necessary under OSX/cocoa.
            if wx.Platform == '__WXGTK__':
                w.Bind(wx.EVT_MOUSEWHEEL, scroll)

            # Listen for mouse down events
            # if cells are selectable
            if self.__selectable and not w.AcceptsFocus():
                w.Bind(wx.EVT_LEFT_DOWN, self.__onLeftMouseDown)

            # Attach the row/column indices to the
            # widget - they are used in the
            # __onLeftDown and __onChildFocus methods
            w._wg_row = row
            w._wg_col = col

        if isinstance(widget, wx.Sizer):
            for c in widget.GetChildren():
                initWidget(c.GetWindow())
        else:
            initWidget(widget)


    def __selectCell(self, row, col):
        """Called by the :meth:`__onChildFocus` and :meth:`__onLeftMouseDown`
        methods. Selects the specified row/column, and generates an
        :data:`EVT_WG_SELECT` event.
        """

        if   self.__selectable == 'rows':    col = -1
        elif self.__selectable == 'columns': row = -1

        try:
            if not self.SetSelection(row, col):
                return

        except ValueError:
            return

        log.debug('Posting grid select event ({}, {})'.format(row, col))

        # This is a ridiculous workaround to a
        # ridiculous problem. Certain users of
        # the WidgetGrid focus a widget within
        # the grid on a select event. This
        # triggers a call to __onChildFocus.
        # Now, the logic in __onChildFocus
        # should be smart enough to not generate
        # another select event when the selected
        # cell hasn't changed. But under GTK it
        # seems that if keyboard/mouse events
        # occur fast enough, the order in which
        # event objects are passed becomes
        # unstable. This means that we can get
        # caught in a silly infinite focus
        # switching loop.
        #
        # To avoid this, I'm setting a flag here,
        # to tell the __onChildFocus method to
        # do nothing while any grid select event
        # handlers are running.
        self.__ignoreFocus += 1

        event = WidgetGridSelectEvent(row=row, col=col)
        event.SetEventObject(self)
        wx.PostEvent(self, event)

        def resetFocus():
            self.__ignoreFocus -= 1

        wx.CallAfter(resetFocus)


    def __onChildFocus(self, ev):
        """If this ``WidgetGrid`` is selectable, this method is called when a
        widget in the grid gains focus. Ensures that the containing cell is
        selected.
        """

        # We explicitly do not call ev.Skip
        # because otherwise the native wx code
        # will automatically scroll to show the
        # focused child, which potentially
        # interferes with application code. The
        # __selectCell method calls SetSelection
        # which then calls __scrollTo, which
        # makes sure that the selected cell is
        # visible.

        # See silliness in __selectCell
        if self.__ignoreFocus > 0:
            return

        gridWidget = ev.GetEventObject()

        row = None
        col = None

        # The event source may be a child of the
        # widget that was added to the grid, so we
        # search up the hierarchy to find the
        # parent that has row and column attributes
        while gridWidget is not None:

            if hasattr(gridWidget, '_wg_row'):
                row = gridWidget._wg_row
                col = gridWidget._wg_col
                break
            else:
                gridWidget = gridWidget.GetParent()

        if row is not None and col is not None:
            log.debug('Focus on cell {}'.format((row, col)))
            self.__selectCell(row, col)


    def __onLeftMouseDown(self, ev):
        """If this ``WidgetGrid`` is selectable, this method is called
        whenever an left mouse down event occurs on an item in the grid.
        """

        widget = ev.GetEventObject()
        row    = widget._wg_row
        col    = widget._wg_col

        if not widget.AcceptsFocus():
            self.SetFocusIgnoringChildren()

        log.debug('Left mouse down on cell {}'.format((row, col)))

        self.__selectCell(row, col)


    def __onKeyboard(self, ev):
        """If the :data:`WG_KEY_NAVIGATION` style is enabled, this method is
        called when the user pushes a key while this ``WidgetGrid`` has focus.
        It changes the currently selected cell, row, or column.
        """
        ev.ResumePropagation(wx.EVENT_PROPAGATE_MAX)

        key = ev.GetKeyCode()

        up    = self.__upKey
        down  = self.__downKey
        left  = self.__leftKey
        right = self.__rightKey

        log.debug('Keyboard event ({})'.format(key))

        # ignore modified keypresses, and all
        # keypresses that are not arrows
        if ev.HasModifiers() or (key not in (up, down, left, right)):
            ev.Skip()
            return

        # if up/down, but we can't select rows
        if key in (up, down) and self.__selectable == 'columns':
            ev.Skip()
            return

        # if left/right, but we can't select columns
        if key in (left, right) and self.__selectable == 'rows':
            ev.Skip()
            return

        row, col = self.__selected

        if   key == up:    row -= 1
        elif key == down:  row += 1
        elif key == left:  col -= 1
        elif key == right: col += 1

        log.debug('Keyboard nav on cell {} ' '(new cell: '
                  '{})'.format(self.__selected, (row, col)))

        self.__selectCell(row, col)


    def GetSelection(self):
        """Returns the currently selected item, as a tuple of ``(row, col)``
        indices. If an entire row has been selected, the ``col`` index  will
        be -1, and vice-versa. If nothing is selected, ``None`` is returned.
        """
        return self.__selected


    def SetSelection(self, row, col):
        """Select the given item. A :exc:`ValueError` is raised if the
        selection is invalid.

        :arg row: Row index of item to select. Pass in -1 to select a whole
                  column.

        :arg col: Column index of item to select. Pass in -1 to select a whole
                  row.

        :returns: ``True`` if the selected item was changed, ``False``
                  otherwise.
        """

        if self.__selected == (row, col):
            return False

        nrows, ncols = self.GetGridSize()

        if self.__selectable == 'rows':

            if col != -1 or row < 0 or row >= nrows:
                raise ValueError('Invalid row: {}'.format(row))

        elif self.__selectable == 'columns':

            if row != -1 or col < 0 or col >= ncols:
                raise ValueError('Invalid column: {}'.format(col))

        elif self.__selectable == 'cells':

            if row < 0 or row >= nrows or col < 0 or col >= ncols:
                raise ValueError('Invalid cell: {}, {}'.format(row, col))

        if self.__selected is not None:

            lsrow, lscol = self.__selected
            self.__selected = None
            self.__select(lsrow, lscol, self.__selectable, False)

        self.__selected = row, col

        self.__select(row, col, self.__selectable, True)
        self.__scrollTo(row, col)

        return True


    def __scrollTo(self, row, col):
        """If scrolling is enabled, this method makes sure that the specified
        row/column is visible.
        """

        # No scrolling
        if not (self.__hscroll or self.__vscroll):
            return

        if row == -1: row = 0
        if col == -1: col = 0

        # We're assuming that the
        # scroll rate is in pixels
        startx,    starty    = self                    .GetViewStart()
        sizex,     sizey     = self                    .GetClientSize()
        posx,      posy      = self.__widgets[row][col].GetPosition()
        widgSizex, widgSizey = self.__widgets[row][col].GetSize()

        # Take into account the size
        # of the widget in the cell
        sizex -= widgSizex
        sizey -= widgSizey

        # take into account the drag
        # panel if it is visible
        if self.__draggable:
            sizey -= self.__dragPanel.GetClientSize().GetHeight()

        scrollx = startx
        scrolly = starty

        # Figure out if the widget is
        # currently visible and, if
        # not, scroll so that it is.
        if   posx < startx:         scrollx = posx
        elif posx > startx + sizex: scrollx = posx - sizex

        if   posy < starty:         scrolly = posy
        elif posy > starty + sizey: scrolly = posy - sizey

        if scrollx != startx or scrolly != starty:
            self.Scroll(scrollx, scrolly)


    def __select(self, row, col, selectType, select=True):
        """Called by the :meth:`SetSelection` method. Sets the background
        colour of the specified row/column to the selection colour, or the
        default colour.

        :arg row:        Row index. If -1, the colour of the entire column is
                         toggled.

        :arg col:        Column index. If -1, the colour of the entire row is
                         toggled.

        :arg selectType: Either ``'rows'``, ``'columns'``, or ``'cells'``.

        :arg select:     If ``True``, the item background colour is set to the
                         selected colour, otherwise it is set to its default
                         colour.
        """

        nrows, ncols = self.GetGridSize()

        if row == -1 and col == -1:
            return

        if nrows == 0 or ncols == 0:
            return

        if row == -1:
            rows = range(nrows)
            cols = [col] * nrows

        elif col == -1:
            rows = [row] * ncols
            cols = range(ncols)

        else:
            rows = [row]
            cols = [col]

        for row, col in zip(rows, cols):

            if   select:  colour = self.__selectedColour
            elif row % 2: colour = self.__oddColour
            else:         colour = self.__evenColour

            container = self.__widgets[   row][col]
            widget    = self.__widgetRefs[row][col]
            self.__setBackgroundColour(container, colour)
            self.__setBackgroundColour(widget,    colour)
            widget   .Refresh()
            container.Refresh()


    def ShowRowLabels(self, show=True):
        """Shows/hides the grid row labels.  The :meth:`Refresh` method must
        be called afterwards for this method to take effect.
        """
        self.__showRowLabels = show


    def ShowColLabels(self, show=True):
        """Shows/hides the grid column labels. The :meth:`Refresh` method must
        be called afterwards for this method to take effect.
        """
        self.__showColLabels = show


    def SetRowLabel(self, row, label):
        """Sets a label for the specified row.

        Raises an :exc:`IndexError` if the row is invalid.
        """
        if row < 0 or row >= self.__nrows:
            raise IndexError('Row {} out of bounds ({})'.format(
                row, self.__nrows))

        self.__rowLabels[row][1].SetLabel(label)


    def SetColLabel(self, col, label):
        """Sets a label for the specified column.

        Raises an :exc:`IndexError` if the column is invalid.
        """
        if col < 0 or col >= self.__ncols:
            raise IndexError('Column {} out of bounds ({})'.format(
                col, self.__ncols))

        self.__colLabels[col][1].SetLabel(label)


    def SetRowLabels(self, labels):
        """Sets the label for every row.
        """
        if len(labels) != self.__nrows:
            raise ValueError('Wrong number of row labels ({} != {})'.format(
                len(labels), self.__nrows))

        for i, label in enumerate(labels):
            self.__rowLabels[i][1].SetLabel(label)


    def SetColLabels(self, labels):
        """Sets the label for every column.
        """
        if len(labels) != self.__ncols:
            raise ValueError('Wrong number of column labels ({} != {})'.format(
                len(labels), self.__ncols))

        for i, label in enumerate(labels):
            self.__colLabels[i][1].SetLabel(label)


    def GetRowLabel(self, row):
        """Return the label of the specified ``row``. """
        return self.__rowLabels[row][1].GetLabel()


    def GetColLabel(self, col):
        """Return the label of the specified ``column``. """
        return self.__colLabels[col][1].GetLabel()


    def GetRowLabels(self):
        """Return all row labels. """
        return [self.__rowLabels[i][1].GetLabel() for i in range(self.__nrows)]


    def GetColLabels(self):
        """Return all column labels. """
        return [self.__colLabels[i][1].GetLabel() for i in range(self.__ncols)]


    def ReorderColumns(self, order):
        """Re-orders the grid columns according to the given sequence
        of column indices.

        A call to this method must be followed by a call to :meth:`Refresh`.

        :arg order: Sequence of column indices (starting from 0) specifying
                    the new column ordering.
        """
        if list(sorted(order)) != list(range(self.__ncols)):
            raise ValueError('Invalid column order (ncols: {}): {}'.format(
                self.__ncols, order))

        self.__colLabels = [self.__colLabels[i] for i in order]

        for rowi in range(self.__nrows):
            widgets    = self.__widgets[   rowi]
            widgetRefs = self.__widgetRefs[rowi]
            self.__widgets[   rowi] = [widgets[i]    for i in order]
            self.__widgetRefs[rowi] = [widgetRefs[i] for i in order]


    def __onColumnLabelMouseDown(self, ev):
        """Called on mouse down events on a column label. """

        ev.Skip()

        lbl = ev.GetEventObject()
        col = self.GetColumn(lbl)

        if col == -1 or (self.__dragLimit > -1 and col > self.__dragLimit):
            return

        self.__dragStartCol   = col
        self.__dragCurrentCol = col

        self.__colLabels[col][0].SetBackgroundColour(self.__dragColour)
        self.__colLabels[col][1].SetBackgroundColour(self.__dragColour)

        # on macOS, an explicit refresh is
        # needed on the backing wx.Panel to
        # ensure that its background colour
        # is updated
        self.__colLabels[col][0].Refresh()


    def __getColumnDragPosition(self):
        """Called during a column drag/drop.

        Returns the current insert index of the dragged column, if it were
        to be dropped now.
        """

        startcol = self.__dragStartCol
        atpos    = wx.FindWindowAtPointer()[0]
        atpos    = self.__getCellPanel(atpos)
        endcol   = self.GetColumn(atpos)

        if endcol == -1 or startcol == endcol:
            return endcol

        # Figure out which side of the drop
        # column to place the dragged column
        # (i.e. on either its left or right
        # side).
        lenx   = atpos.GetSize().GetWidth()
        posx   = atpos.ScreenToClient(wx.GetMouseState().GetPosition()).x
        posx   = posx / lenx
        endcol = int(round(endcol + posx))

        # Clip to the drag limit column if
        # it is set
        if self.__dragLimit > -1:
            endcol = min(endcol, self.__dragLimit + 1)

        return endcol


    def __onColumnLabelMouseDrag(self, ev):
        """Called during a column drag. Updates the marker location on
        the drag panel.
        """
        if self.__dragStartCol is None:
            return

        startcol   = self.__dragStartCol
        lastcol    = self.__dragCurrentCol
        currentcol = self.__getColumnDragPosition()
        panel      = self.__dragPanel

        # mouse is off the grid, or
        # current drop position would
        # not move the column
        if currentcol == -1         or \
           startcol   == currentcol or \
           startcol   == currentcol - 1:
            self.__dragCurrentCol = None
            panel.ClearBackground()
            return

        # No change since last draw
        if lastcol == currentcol:
            return

        self.__dragCurrentCol = currentcol

        self.__dragPanel.Refresh()


    def __onColumnLabelMouseUp(self, ev):
        """Called on the mouse up event at the end of a column drag.

        Re-orders the grid columns.
        """

        if self.__dragStartCol is None:
            return

        # The start column was saved in the
        # mousedown handler. Figure out the
        # column that mouseup occurred in.
        startcol              = self.__dragStartCol
        endcol                = self.__getColumnDragPosition()
        self.__dragStartCol   = None
        self.__dragCurrentCol = None
        self.__dragPanel.ClearBackground()
        self.__dragPanel.Refresh()

        self.__colLabels[startcol][0].SetBackgroundColour(self.__labelColour)
        self.__colLabels[startcol][1].SetBackgroundColour(self.__labelColour)
        self.__colLabels[startcol][0].Refresh()

        if endcol == -1 or startcol == endcol:
            return

        # Offset the insertion index if
        # the starting column is before
        # it, as it gets removed before
        # being re-added
        if startcol < endcol:
            endcol = endcol - 1

        # Generate a new column ordering
        order = list(range(self.__ncols))
        order.pop(startcol)

        order.insert(endcol, startcol)

        self.ReorderColumns(order)
        self.Refresh()

        event = WidgetGridReorderEvent(order=order)
        event.SetEventObject(self)
        wx.PostEvent(self, event)


    def __dragPanelPaint(self, ev):
        """Paints the current column drop location on the drag panel. """

        currentcol = self.__dragCurrentCol
        dc         = wx.PaintDC(self.__dragPanel)

        if not dc.IsOk():
            return

        if currentcol is None:
            return

        dwidth, dheight = dc.GetSize().Get()

        if dwidth == 0 or dheight == 0:
            return

        if currentcol == self.__ncols:
            szitem = self.__gridSizer.GetItem(self.__colLabels[-1][0])
            xpos   = szitem.GetPosition()[0] + szitem.GetSize()[0]

        else:
            szitem = self.__gridSizer.GetItem(self.__colLabels[currentcol][0])
            xpos   = szitem.GetPosition()[0]

        xpos -= 0.5 * self.__dragIcon.GetSize()[0]

        self.__dragCurrentCol = currentcol
        dc.Clear()
        dc.DrawBitmap(self.__dragIcon, int(xpos), 0, False)


WG_SELECTABLE_CELLS = 1
"""If this style is enabled, individual cells can be selected. """


WG_SELECTABLE_ROWS = 2
"""If this style is enabled, whole rows can be selected. """


WG_SELECTABLE_COLUMNS = 4
"""If this style is enabled, whole columns can be selected. """


WG_KEY_NAVIGATION = 8
"""If this style is enabled along with one of the ``*_SELECTABLE_*`` styles,
the user may use the keyboard to navigate between cells, rows, or columns.
"""


WG_DRAGGABLE_COLUMNS = 16
"""If this style is enabled, column names can be dragged with the mouse to
re-order them.
"""


_WidgetGridSelectEvent,  _EVT_WG_SELECT  = wxevent.NewEvent()
_WidgetGridReorderEvent, _EVT_WG_REORDER = wxevent.NewEvent()


EVT_WG_SELECT = _EVT_WG_SELECT
"""Identifier for the :data:`WidgetGridSelectEvent`. """


EVT_WG_REORDER = _EVT_WG_REORDER
"""Identifier for the :data:`WidgetGridReorderEvent`. """


WidgetGridSelectEvent = _WidgetGridSelectEvent
"""Event generated when an item in a ``WidgetGrid`` is selected. A
``WidgetGridSelectEvent`` has the following attributes:

 - ``row`` Row index of the selected item. If -1, an entire column
   has been selected.

 - ``col`` Column index of the selected item. If -1, an entire row
   has been selected.
"""


WidgetGridReorderEvent = _WidgetGridReorderEvent
"""Event generated when the columns in a ``WidgetGrid`` are reordered. A
``WidgetGridReorderEvent`` has the following attributes:

 - ``order`` The new column order.
"""


TRIANGLE_ICON = b'''
iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABmJLR0QA/wD/AP+gvaeTAAAA
CXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4wIFDQoeGSImZAAAACZpVFh0Q29tbWVudAAA
AAAAQ3JlYXRlZCB3aXRoIEdJTVAgb24gYSBNYWOV5F9bAAAAZklEQVQY053PwQmEUAyE4U8F
tw5L8KAlbksetgIrsBLRy8O9RBDxCTowhyTDH4YH+iFhyzhFRn8T2t3v1DFDTXEDJdobWouy
imHBig51AGZ8MWAtTsW201xctf+gObxsYpfVFH6nP5vqKwqbBq3zAAAAAElFTkSuQmCC
'''.strip().replace(b'\n', b'')
"""Icon used as the drop marker when columns are being re-ordered by
mouse drag.
"""



#!/usr/bin/env python
#
# setup.py - setuptools configuration for installing the fsleyes-widgets
# package.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


from __future__ import print_function

import os.path as op
import            shutil

from setuptools import setup
from setuptools import find_packages
from setuptools import Command


basedir = op.dirname(__file__)

# Dependencies are listed in requirements.txt
with open(op.join(basedir, 'requirements.txt'), 'rt') as f:
    install_requires = [l.strip() for l in f.readlines()]

packages = find_packages(include=('fsleyes_widgets', 'fsleyes_widgets.*'))

# Extract the vesrion number from fsleyes_widgets/__init__.py
version = {}
with open(op.join(basedir, "fsleyes_widgets", "__init__.py"), 'rt') as f:
    for line in f:
        if line.startswith('__version__'):
            exec(line, version)
            break
version = version.get('__version__')

with open(op.join(basedir, 'README.rst'), 'rt') as f:
    readme = f.read()


class doc(Command):
    """Build the API documentation. """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        docdir  = op.join(basedir, 'doc')
        destdir = op.join(docdir, 'html')

        if op.exists(destdir):
            shutil.rmtree(destdir)

        print('Building documentation [{}]'.format(destdir))

        import sphinx.cmd.build as sphinx_build

        try:
            import unittest.mock as mock
        except Exception:
            import mock

        mockobj       = mock.MagicMock()
        mockedModules = open(op.join(docdir, 'mock_modules.txt')).readlines()
        mockedClasses = open(op.join(docdir, 'mock_classes.txt')).readlines()

        mockedModules = [l.strip()   for l in mockedModules]
        mockedClasses = [l.strip()   for l in mockedClasses]
        mockedModules = {m : mockobj for m in mockedModules}

        patches = [mock.patch.dict('sys.modules', **mockedModules)] + \
                  [mock.patch('wx.lib.newevent.NewEvent',
                              return_value=(mockobj, mockobj))] + \
                  [mock.patch(c, object) for c in mockedClasses]

        [p.start() for p in patches]
        sphinx_build.main([docdir, destdir])
        [p.stop() for p in patches]


setup(

    name='fsleyes-widgets',
    version=version,
    description='A collection of wxPython widgets used by FSLeyes',
    long_description=readme,
    long_description_content_type='text/x-rst',
    url='https://git.fmrib.ox.ac.uk/fsl/fsleyes/widgets',
    author='Paul McCarthy',
    author_email='pauldmccarthy@gmail.com',
    license='Apache License Version 2.0',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules'],

    packages=packages,
    install_requires=install_requires,
    test_suite='tests',

    cmdclass={
        'doc' : doc
    }
)



#!/usr/bin/env python
#
# __init__.py - Small utility modules for doing random things.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains a collection of small utility modules for doing
random things. A few functions are also defined at the package level:

.. autosummary::
   :nosignatures:

    wxVersion
    wxFlavour
    wxPlatform
    frozen
    canHaveGui
    haveGui
    inSSHSession
    inVNCSession
    isalive


This package can be imported, and all of its functions called (with the
exception of :func:`isalive`), without ``wx`` being installed.
"""


import functools as ft
import              os
import              sys
import              warnings


WX_PYTHON  = 1
"""Constant returned by the :func:`wxFlavour` function, indicating that
wxPython 3.0.2.0 or older is being used.
"""


WX_PHOENIX = 2
"""Constant returned by the :func:`wxFlavour` function, indicating that
wxPython/Phoenix (>= 3.0.3) is being used.
"""


WX_UNKNOWN = 0
"""Constant returned by the :func:`wxPlatform` function, indicating that an
unknown wx version is being used.
"""


WX_MAC_COCOA = 1
"""Constant returned by the :func:`wxPlatform` function, indicating that an
OSX cocoa wx version is being used.
"""


WX_MAC_CARBON = 2
"""Constant returned by the :func:`wxPlatform` function, indicating that an
OSX carbon wx version is being used.
"""


WX_GTK = 3
"""Alias for :attr:`WX_GTK2` - will be removed in a future version of
``fsleyes-widgets``.
"""


WX_GTK2 = 3
"""Constant returned by the :func:`wxPlatform` function, indicating that a
Linux/GTK3 wx version is being used.
"""


WX_GTK3 = 4
"""Constant returned by the :func:`wxPlatform` function, indicating that a
Linux/GTK3 wx version is being used.
"""


def wxVersion():
    """Return a string containing the "major.minor.patch" version of wxPython
    that is installed, or ``None`` if wxPython is not installed, or its version
    cannot be determined.
    """

    try:
        import wx

        # Only consider the first three components
        # (e.g. ignore the "post2" in "4.0.7.post2")
        version = [int(v) for v in wx.__version__.split('.')[:3]]
        return '.'.join([str(v) for v in version])

    except Exception:
        return None


def wxFlavour():
    """Determines the version of wxPython being used. Returns ``WX_PYTHON``,
    ``WX_PHOENIX``, ``or ``None`` if wxPython does not appear to be installed.
    """

    try:
        import wx
    except ImportError:
        return None

    pi        = [t.lower() for t in wx.PlatformInfo]
    isPhoenix = False

    for tag in pi:
        if 'phoenix' in tag:
            isPhoenix = True
            break

    if isPhoenix: return WX_PHOENIX
    else:         return WX_PYTHON


def wxversion():
    """Deprecated - use ``wxFlavour``instead. """
    warnings.warn('wxversion is deprecated - use wxFlavour instead.',
                  category=DeprecationWarning,
                  stacklevel=2)
    return wxFlavour()


def wxPlatform():
    """Returns one of :data:`WX_UNKNOWN` :data:`WX_MAC_COCOA`,
    :data:`WX_MAC_CARBON`, :data:`WX_GTK2`, or :data:`WX_GTK3`, indicating the
    wx platform, or ``None`` if wxPython does not appear to be installed.
    """

    try:
        import wx
    except ImportError:
        return None

    pi = [t.lower() for t in wx.PlatformInfo]

    if   any(['cocoa'  in p for p in pi]): plat = WX_MAC_COCOA
    elif any(['carbon' in p for p in pi]): plat = WX_MAC_CARBON
    elif any(['gtk3'   in p for p in pi]): plat = WX_GTK3
    elif any(['gtk2'   in p for p in pi]): plat = WX_GTK2
    elif any(['gtk'    in p for p in pi]): plat = WX_GTK2
    else:                                  plat = WX_UNKNOWN

    return plat


def frozen():
    """``True`` if we are running in a compiled/frozen application
    (e.g. pyinstaller, py2app), ``False`` otherwise.
    """
    return getattr(sys, 'frozen', False)


@ft.lru_cache()
def canHaveGui():
    """Return ``True`` if a display is available, ``False`` otherwise. """

    # We cache this because calling the
    # IsDisplayAvailable function will cause the
    # application to steal focus under OSX!
    try:
        import wx
        return wx.App.IsDisplayAvailable()
    except ImportError:
        return False


def haveGui():
    """``True`` if we are running with a GUI, ``False`` otherwise.

    This currently equates to testing whether a display is available
    (see :meth:`canHaveGui`) and whether a ``wx.App`` exists. It
    previously also tested whether an event loop was running, but this
    is not compatible with execution from IPython/Jupyter notebook, where
    the event loop is called periodically, and so is not always running.
    """
    try:
        import wx
    except ImportError:
        return False

    app = wx.GetApp()
    # TODO Previously this conditional
    #      also used app.IsMainLoopRunning()
    #      to check that the wx main loop
    #      was running. But this doesn't
    #      suit situations where a non-main
    #      event loop is running, or where
    #      the mainloop is periodically
    #      started and stopped (e.g. when
    #      the event loop is being run by
    #      IPython).
    #
    #      In c++ wx, there is the
    #      wx.App.UsesEventLoop method, but
    #      this is not presently exposed to
    #      Python code (and wouldn't help
    #      to detect the loop start/stop
    #      scenario).
    #
    #      So this constraint has been
    #      (hopefully) temporarily relaxed
    #      until I can think of a better
    #      solution.
    return (canHaveGui() and app is not None)


def inSSHSession():
    """Return ``True`` if this application appears to be running over an SSH
    session, ``False`` otherwise.
    """
    sshVars = ['SSH_CLIENT', 'SSH_TTY']
    return any(s in os.environ for s in sshVars)


def inVNCSession():
    """Returns ``True`` if this application appears to be running over a VNC (or
    similar) session, ``False`` otherwise. Currently, the following remote
    desktop environments are detected:

      - VNC
      - x2go
      - NoMachine
    """
    vncVars = ['VNCDESKTOP', 'X2GO_SESSION', 'NXSESSIONID']
    return any(v in os.environ for v in vncVars)


def isalive(widget):
    """Returns ``True`` if the given ``wx.Window`` object is "alive" (i.e.  has
    not been destroyed), ``False`` otherwise. Works in both wxPython and
    wxPython/Phoenix.

    .. warning:: Don't try to test whether a ``wx.MenuItem`` has been
                 destroyed, as it will probably result in segmentation
                 faults. Check the parent ``wx.Menu`` instead.
    """

    import wx

    wxver = wxFlavour()

    if wxver == WX_PHOENIX:
        excTypes = (RuntimeError,)
    else:
        excTypes = (wx.PyDeadObjectError,)

    try:
        if hasattr(widget, 'IsBeingDeleted'):
            return not widget.IsBeingDeleted()
        else:
            widget.GetParent()
            return True

    except excTypes:
        return False



#!/usr/bin/env python
#
# colourbutton.py - A button which allows the user to select a colour.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ColourButton` class, a button which
allows the user to select a RGBA colour.
"""


import wx
import wx.lib.newevent as wxevent

import fsleyes_widgets as fw


class ColourButton(wx.Button):
    """A :class:`wx.Button` which allows the user to select a colour.

    The currently selected colour is displayed as a bitmap on the button.
    When the user presses the button, a :class:`wx.ColourDialog` is displayed,
    allowing the user to change the colour. When the user does so, a
    :data:`EVT_COLOUR_BUTTON_EVENT` is generated.

    This class provides an alternative to the :class:`wx.ColourPickerCtrl`,
    which is a bit inflexible w.r.t. sizing/automatic resizing.
    """

    def __init__(self, parent, size=None, colour=None):
        """Create a ``ColourButton``.

        :arg parent: A :mod:`wx` parent window.

        :arg size:   A tuple containing the ``(width, height)`` of the
                     colour bitmap in pixels. Defaults to ``(32, 32)``.

        :arg colour: Initial colour. Defaults to black.
        """

        if size   is None: size   = (32, 32)
        if colour is None: colour = (0, 0, 0, 255)

        style = wx.BU_EXACTFIT | wx.BU_NOTEXT

        wx.Button.__init__(self, parent, style=style)

        # Under wxPython-phoenix, setting
        # label='' results in "Button".
        if fw.wxFlavour() == fw.WX_PHOENIX:
            self.SetLabel(' ')

        self.__size = size
        self.__bmp  = None

        self.Bind(wx.EVT_BUTTON, self.__onClick)

        self.SetValue(colour)
        self.SetMinSize(self.GetBestSize())


    def GetValue(self):
        """Return the current colour, as a tuple of ``(r, g, b, a)`` values,
        each in the range ``[0 - 255]``.
        """
        return self.__colour


    def SetValue(self, colour):
        """Sets the current colour to the specified ``colour``."""

        if len(colour) not in (3, 4):
            raise ValueError('Invalid RGB[A] colour: {}'.format(colour))

        if len(colour) == 3:
            colour = list(colour) + [255]

        if any([v < 0 or v > 255 for v in colour]):
            raise ValueError('Invalid RGBA colour: {}'.format(colour))

        self.__updateBitmap(colour)
        self.__colour = colour


    def __updateBitmap(self, colour):
        """Called when the colour is changed. Updates the bitmap shown
        on the button.
        """

        import numpy as np

        w, h = self.__size
        data = np.zeros((w, h, 4), dtype=np.uint8)

        data[:, :] = colour

        if fw.wxFlavour() == fw.WX_PHOENIX:
            self.__bmp = wx.Bitmap.FromBufferRGBA(w, h, data)
        else:
            self.__bmp = wx.BitmapFromBufferRGBA( w, h, data)

        self.SetBitmap(self.__bmp)


    def __onClick(self, ev):
        """Called when this ``ColourButton`` is pressed.

        Displays a :class:`wx.ColourDialog` allowing the user to select a new
        colour.
        """

        colourData = wx.ColourData()
        colourData.SetColour(self.__colour)

        dlg = wx.ColourDialog(self.GetTopLevelParent(), colourData)

        if dlg.ShowModal() != wx.ID_OK:
            return

        newColour = dlg.GetColourData().GetColour()
        newColour = [newColour.Red(),
                     newColour.Green(),
                     newColour.Blue(),
                     newColour.Alpha()]

        self.SetValue(newColour)

        wx.PostEvent(self, ColourButtonEvent(colour=newColour))


_ColourButtonEvent, _EVT_COLOUR_BUTTON_EVENT = wxevent.NewEvent()


EVT_COLOUR_BUTTON_EVENT = _EVT_COLOUR_BUTTON_EVENT
"""Identifier for the :data:`ColourButtonEvent`. """


ColourButtonEvent = _ColourButtonEvent
"""Event emitted by a ``ColourButton`` when the user changes the selected
colour.
"""



#!/usr/bin/env python
#
# test_status.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import time

from unittest import mock
import pytest

import fsleyes_widgets.utils.status as status


class MockTarget(object):

    def __init__(self):
        self.msg = None

    def __call__(self, msg):
        self.msg = msg


def test_update():

    # Test without a target
    status.setTarget(None)
    status.update('Status')

    target = MockTarget()

    status.setTarget(target)

    assert target.msg is None
    status.update('Status1', None)
    assert target.msg == 'Status1'
    status.update('Status2', None)
    assert target.msg == 'Status2'


def test_update_timeout():

    status.setTarget(None)
    status.clearStatus()

    target = MockTarget()
    status.setTarget(target)

    # Test that the message gets cleared
    status.update('Status1', 0.25)
    assert target.msg == 'Status1'
    time.sleep(0.5)
    assert target.msg == ''

    # If a timed update is followed by an untimed
    # update, the untimed one should persist
    status.update('Status2', 0.25)
    status.update('Status3', None)
    time.sleep(0.5)
    assert target.msg == 'Status3'


def test_clearStatus():
    target = MockTarget()
    status.setTarget(target)

    status.update('Status1', None)
    assert target.msg == 'Status1'
    status.clearStatus()
    assert target.msg == ''

    status.update('Status1', 0.25)
    assert target.msg == 'Status1'
    status.clearStatus()
    assert target.msg == ''


def test_ClearThread_die():
    target = MockTarget()
    status.setTarget(target)

    # make sure the clearthread is running
    status.update('Status1', 1.0)
    time.sleep(1.1)

    # and can be killed
    status._clearThread.die()
    status._clearThread.clear(0.1)
    status._clearThread.join()
    status._clearThread = None

    # and then gets restarted again
    status.update('Status1', 0.25)
    assert target.msg == 'Status1'
    time.sleep(0.5)
    assert target.msg == ''






def test_reportError():

    class MockWX(object):

        ICON_ERROR = 0
        OK         = 2

        def MessageBox(self, msg, title, flags):
            self.msg   = msg
            self.title = title

    class MyException(Exception):
        def __str__(self):
            return 'MyException'

    wx  = MockWX()
    exc = MyException()

    with mock.patch.dict('sys.modules', {'wx' : wx}):
        status.reportError('Error title',
                           'Error message',
                           exc)

    assert wx.title == 'Error title'
    assert wx.msg.startswith('Error message')
    assert wx.msg.endswith(str(exc))

    with mock.patch.dict('sys.modules', {'wx' : None}), \
         mock.patch('fsleyes_widgets.utils.status.log.error') as log:
        status.reportError('Error title', 'Error message', exc)
        log.assert_called_once()


def test_reportIfError_noError():

    with mock.patch('fsleyes_widgets.utils.status.reportError') as func:
        with status.reportIfError('title', 'message'):
            pass
        func.assert_not_called()


def test_reportIfError_error_noraise():

    title = 'title'
    msg   = 'msg'
    exc   = Exception()

    # report=True
    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr:
        with status.reportIfError(title, msg, raiseError=False):
            raise exc
        reperr.assert_called_once_with(title, msg, exc)

    # report=False
    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr, \
         mock.patch('fsleyes_widgets.utils.status.log.error') as log:
        with status.reportIfError(title, msg, raiseError=False, report=False):
            raise exc
        reperr.assert_not_called()
        log.assert_called_once()


def test_reportIfError_error_raise():

    title = 'title'
    msg   = 'msg'
    exc   = Exception()

    # report=True
    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr:
        with pytest.raises(Exception):
            with status.reportIfError(title, msg):
                raise exc
        reperr.assert_called_once_with(title, msg, exc)

    # report=False
    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr, \
         mock.patch('fsleyes_widgets.utils.status.log.error') as log:

        with pytest.raises(Exception):
            with status.reportIfError(title, msg, report=False):
                raise exc
        reperr.assert_not_called()
        log.assert_called_once()


def test_reportErrorDecorator():

    title = 'title'
    msg   = 'msg'
    exc   = Exception()

    @status.reportErrorDecorator(title, msg)
    def errfunc():
        raise exc

    @status.reportErrorDecorator(title, msg)
    def noerrfunc():
        pass

    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr:
        noerrfunc()
        reperr.assert_not_called()

    with mock.patch('fsleyes_widgets.utils.status.reportError') as reperr:
        with pytest.raises(Exception):
            errfunc()
        reperr.assert_called_with(title, msg, exc)
